import ij.*;
import ij.process.*;
import ij.gui.*;
import ij.io.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import ij.plugin.*;
import ij.plugin.frame.*;

import ij.util.*;
import ij.plugin.DICOM;
import java.util.*;
import ij.measure.*;


    /**
      This plugin implements the Multi/Quad-Modal Anomalous Diffusion regression as documented in .

      @author Fred Damen <fred@damen.org> <fdamen1@uic.edu>

      Version History:
      2020-04-01: First Release
    */

public class Diffusion_MultiModal implements PlugIn, MouseListener, 
                                                     MouseMotionListener, 
                                                     KeyListener,  
                                                     WindowListener {

   ImagePlus dimp = null;
   ImagePlus[] Rimps   = null;
   Point loc = null;
   int dx=0, dy=0;
   int pww=0, pwh=0;

   String[] models = new String[]{"Multimodal","ADC","IVIM","Stretch","IVIM-Stretch"};
   String model = models[0];

   static double NOISEFLOOR = 5F;  //0.01F; //

   public void run(String arg) {
      try {
         int ugh = 0;
         loc = (Point)IJ.getProperty("PreferedLocation");
         ImagePlus imp = WindowManager.getCurrentImage();
         if (imp != null) {
            loc = imp.getWindow().getLocation();
            GenericDialog gd = new GenericDialog("Diffusion MultiModal");
            if (loc != null) gd.setLocation(loc.x, loc.y);
            gd.addMessage("Process "+imp.getTitle());
            if (imp.getType() != ImagePlus.GRAY32) {
               Font font = gd.getFont();
               if (font == null) font = new Font("Comic Sans MS", Font.BOLD, 20);
               gd.addMessage("is not 32-bit", font.deriveFont(Font.BOLD,font.getSize()+4),Color.red);
               }
            gd.addRadioButtonGroup("Model:", models, models.length, 1, model);
            gd.addMessage("Version 1");
            gd.showDialog();
            if (gd.wasCanceled()) return;
            loc = gd.getLocation();
            //IJ.setProperty("PreferedLocation",loc);
            model = gd.getNextRadioButton();
            }
	 else {
            IJ.error("Diffusion_Multimodal: no open image");
            return;
            }

         dimp = imp;

         if (loc == null) loc = imp.getWindow().getLocation();
         if (loc != null) ImageWindow.setNextLocation(loc.x, loc.y);
         Dimension impWinSize = imp.getWindow().getSize();
         dx = (int)(0.9 * (float)impWinSize.width);
         dy = (int)(0.9 * (float)impWinSize.height);

         loc.x = loc.x + impWinSize.width;

              if (model.equals("Multimodal"))   processAndDisplay(imp);
         else if (model.equals("ADC"))          processADCAndDisplay(imp);
         else if (model.equals("IVIM"))         processIVIMAndDisplay(imp);
         else if (model.equals("Stretch"))      processStretchAndDisplay(imp);
         else if (model.equals("IVIM-Stretch")) processIVIMStretchAndDisplay(imp);
         else IJ.error("Unknown model '"+model+"'");
         }
      catch(Exception e) {
         StringWriter sw = new StringWriter();
         e.printStackTrace(new PrintWriter(sw));
         IJ.error("Diffusion_Multimodal.\n\n"+sw.toString());
         }
      }

//////////////////////////////////////////////////////////////////////////////////
   private void processAndDisplay(ImagePlus imp) {
      Rimps = process(imp);

      Dimension ws = imp.getWindow().getSize();
      Point rloc = new Point(loc.x - 6*ws.width, loc.y);
      ImageWindow.setNextLocation(rloc.x+0*ws.width, rloc.y+0*ws.height); Rimps[0].show();
      ImageWindow.setNextLocation(rloc.x+0*ws.width, rloc.y+1*ws.height); Rimps[1].show();
      ImageWindow.setNextLocation(rloc.x+1*ws.width, rloc.y+0*ws.height); Rimps[2].show();
      ImageWindow.setNextLocation(rloc.x+2*ws.width, rloc.y+0*ws.height); Rimps[3].show();
      ImageWindow.setNextLocation(rloc.x+2*ws.width, rloc.y+1*ws.height); Rimps[4].show();
      ImageWindow.setNextLocation(rloc.x+2*ws.width, rloc.y+2*ws.height); Rimps[5].show();
      ImageWindow.setNextLocation(rloc.x+3*ws.width, rloc.y+0*ws.height); Rimps[6].show();
      ImageWindow.setNextLocation(rloc.x+4*ws.width, rloc.y+0*ws.height); Rimps[7].show();
      ImageWindow.setNextLocation(rloc.x+4*ws.width, rloc.y+1*ws.height); Rimps[8].show();
      ImageWindow.setNextLocation(rloc.x+4*ws.width, rloc.y+2*ws.height); Rimps[9].show();
      for(ImagePlus rimp : Rimps)
         rimp.getWindow().setSize(ws.width,ws.height);
      for(ImagePlus rimp : Rimps)
         rimp.getWindow().getCanvas().fitToWindow();
      //for(ImagePlus rimp : Rimps)
      //   rimp.updateAndRepaintWindow();
      setupPlotting(imp,Rimps);
      }

   public static ImagePlus[] process(ImagePlus imp) {

      ImagePlus mask = (new Duplicator()).run(imp,1,1,1,imp.getNSlices(),1,1);
      IJ.run(mask, "32-bit", "");
      IJ.run(mask,"Convert to Mask", "method=Huang background=Default calculate black");
      IJ.run(mask, "Fill Holes", "stack");
/*
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Dilate", "stack");
      IJ.run(mask, "Dilate", "stack");
      IJ.run(mask, "Dilate", "stack");
*/
      
      ImagePlus[] pimp = F_Project.compute(imp, new F_Project.Compute() {
           @Override
           public double[] compute(double[] x, double[] y) {
              if (x == null) return new double[10];
              return nonlinearFit(x,y);
              }
           }, mask, new ImagePlus[0]);

      double[] ms = meanStddev(pimp[7]);
      ImagePlus filter = pimp[7].duplicate();
      IJ.run(filter, "Macro...", "code=[if(v < "+(ms[0]+4*ms[1])+")v=1; else v=NaN;] stack");
      ImageCalculator ic = new ImageCalculator();
      for(int i=0; i<7; i++)
         ic.run("Multiply 32-bit stack", pimp[i], filter);

      String title = imp.getTitle();

      pimp[0].setTitle("fraction flow "+title);
      pimp[0].setDisplayRange(0,0.3);

      pimp[1].setTitle("pseudo perfusion "+title);
      pimp[1].setDisplayRange(3E-3,50E-3);

      pimp[2].setTitle("fraction Unrestricted "+title);
      pimp[2].setDisplayRange(0,1);

      pimp[3].setTitle("fraction Restricted "+title);
      pimp[3].setDisplayRange(0,1);

      pimp[4].setTitle("Restricted Diffusion "+title);
      pimp[4].setDisplayRange(0E-3,3E-3);

      pimp[5].setTitle("alpha "+title);
      pimp[5].setDisplayRange(0.5,1);

      pimp[6].setTitle("fraction Really Restricted "+title);
      pimp[6].setDisplayRange(0.0,0.3);

      pimp[7].setTitle("chi "+title);
      pimp[7].setDisplayRange(0.00,0.3);

      pimp[8].setTitle("NoiseFloor "+title);

      pimp[9].setTitle("b0 problem "+title);
      pimp[9].setDisplayRange(-1.00,0.00);

      return pimp;
      }

//////////////////////////////////////////////////////////////////////////////////
   public static double[] meanStddev(ImagePlus imp) {
      double sum = 0.0;
      double sumSq = 0.0;
      int n = 0;
      int NS = imp.getNSlices();
      for(int s=0; s<NS; s++)
         for(float[] line : imp.getStack().getProcessor(s+1).getFloatArray())
            for(float v : line)
               if (!Float.isNaN(v)) {
                  n++;
                  sum += v;
                  sumSq += v*v;
                  }
      return new double[]{sum/n,Math.sqrt(sumSq/n - (sum/n)*(sum/n))};
      }

//////////////////////////////////////////////////////////////////////////////////
   private void processADCAndDisplay(ImagePlus imp) {
      Rimps = processADC(imp);

      Dimension ws = imp.getWindow().getSize();
      Point rloc = new Point(loc.x - 4*ws.width, loc.y);
      ImageWindow.setNextLocation(rloc.x+0*ws.width, rloc.y+0*ws.height); Rimps[0].show();
      ImageWindow.setNextLocation(rloc.x+0*ws.width, rloc.y+1*ws.height); Rimps[1].show();
      ImageWindow.setNextLocation(rloc.x+1*ws.width, rloc.y+0*ws.height); Rimps[2].show();
      ImageWindow.setNextLocation(rloc.x+1*ws.width, rloc.y+1*ws.height); Rimps[3].show();
      ImageWindow.setNextLocation(rloc.x+2*ws.width, rloc.y+0*ws.height); Rimps[4].show();
      ImageWindow.setNextLocation(rloc.x+2*ws.width, rloc.y+1*ws.height); Rimps[5].show();
      for(ImagePlus rimp : Rimps)
         rimp.getWindow().setSize(ws.width,ws.height);
      for(ImagePlus rimp : Rimps)
         rimp.getWindow().getCanvas().fitToWindow();
      setupPlotting(imp,Rimps);
      }

   public static ImagePlus[] processADC(ImagePlus imp) {

      ImagePlus mask = (new Duplicator()).run(imp,1,1,1,imp.getNSlices(),1,1);
      IJ.run(mask, "32-bit", "");
      IJ.run(mask,"Convert to Mask", "method=Huang background=Default calculate black");
      IJ.run(mask, "Fill Holes", "stack");
/*
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Dilate", "stack");
      IJ.run(mask, "Dilate", "stack");
      IJ.run(mask, "Dilate", "stack");
*/
      
      ImagePlus[] pimp = F_Project.compute(imp, new F_Project.Compute() {
           @Override
           public double[] compute(double[] x, double[] y) {
              if (x == null) return new double[6];
              double b0 = y[0];
              int[] indexs = new int[]{arrElementNear(x,1000),arrElementNear(x,2000),arrElementNear(x,3000)};
              double[] ADCs = new double[indexs.length];
              double[] rmss = new double[indexs.length];
              for(int i=0; i<indexs.length; i++) {
                 ADCs[i] = -Math.log(y[indexs[i]]/b0)/x[indexs[i]];
                 double[] yy = se(x,1,ADCs[i],1);
                 rmss[i] = Math.sqrt(arrMeanSquareDiff(arrNorm(y),yy));
                 }
              return new double[]{ADCs[0],rmss[0],ADCs[1],rmss[1],ADCs[2],rmss[2]};
              }
           }, mask, new ImagePlus[0]);

      String title = imp.getTitle();

      pimp[0].setTitle("ADC 1000 "+title);
      pimp[0].setDisplayRange(0E-3,3E-3);
      pimp[1].setTitle("chi 1000 "+title);
      pimp[1].setDisplayRange(0.00,0.06);

      pimp[2].setTitle("ADC 2000 "+title);
      pimp[2].setDisplayRange(0E-3,3E-3);
      pimp[3].setTitle("chi 2000 "+title);
      pimp[3].setDisplayRange(0.00,0.06);

      pimp[4].setTitle("ADC 2800 "+title);
      pimp[4].setDisplayRange(0E-3,3E-3);
      pimp[5].setTitle("chi 2800 "+title);
      pimp[5].setDisplayRange(0.00,0.06);

      return pimp;
      }

//////////////////////////////////////////////////////////////////////////////////
   private void processIVIMAndDisplay(ImagePlus imp) {
      Rimps = processIVIM(imp);

      Dimension ws = imp.getWindow().getSize();
      Point rloc = new Point(loc.x - 4*ws.width, loc.y);
      ImageWindow.setNextLocation(rloc.x+0*ws.width, rloc.y+0*ws.height); Rimps[0].show();
      ImageWindow.setNextLocation(rloc.x+0*ws.width, rloc.y+1*ws.height); Rimps[1].show();
      ImageWindow.setNextLocation(rloc.x+1*ws.width, rloc.y+0*ws.height); Rimps[2].show();
      ImageWindow.setNextLocation(rloc.x+1*ws.width, rloc.y+1*ws.height); Rimps[3].show();
      ImageWindow.setNextLocation(rloc.x+2*ws.width, rloc.y+0*ws.height); Rimps[4].show();
      for(ImagePlus rimp : Rimps)
         rimp.getWindow().setSize(ws.width,ws.height);
      for(ImagePlus rimp : Rimps)
         rimp.getWindow().getCanvas().fitToWindow();
      setupPlotting(imp,Rimps);
      }

   public static ImagePlus[] processIVIM(ImagePlus imp) {

      ImagePlus mask = (new Duplicator()).run(imp,1,1,1,imp.getNSlices(),1,1);
      IJ.run(mask, "32-bit", "");
      IJ.run(mask,"Convert to Mask", "method=Huang background=Default calculate black");
      IJ.run(mask, "Fill Holes", "stack");
/*
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Dilate", "stack");
      IJ.run(mask, "Dilate", "stack");
      IJ.run(mask, "Dilate", "stack");
*/
      
      ImagePlus[] pimp = F_Project.compute(imp, new F_Project.Compute() {
           @Override
           public double[] compute(double[] x, double[] y) {
              if (x == null) return new double[5];
              return fitIVIM(x,y);
              }
           }, mask, new ImagePlus[0]);

      String title = imp.getTitle();

      pimp[0].setTitle("fraction pseudo perfusion "+title);
      pimp[0].setDisplayRange(0,0.3);

      pimp[1].setTitle("pseudo perfusion "+title);
      pimp[1].setDisplayRange(3E-3,50E-3);

      pimp[2].setTitle("fraction diffusion "+title);
      pimp[2].setDisplayRange(0,1);

      pimp[3].setTitle("diffusion "+title);
      pimp[3].setDisplayRange(0E-3,3E-3);

      pimp[4].setTitle("chi "+title);
      pimp[4].setDisplayRange(0.00,0.06);

      return pimp;
      }

//////////////////////////////////////////////////////////////////////////////////
   private void processStretchAndDisplay(ImagePlus imp) {
      Rimps = processStretch(imp);

      Dimension ws = imp.getWindow().getSize();
      Point rloc = new Point(loc.x - 3*ws.width, loc.y);
      ImageWindow.setNextLocation(rloc.x+0*ws.width, rloc.y+0*ws.height); Rimps[0].show();
      ImageWindow.setNextLocation(rloc.x+0*ws.width, rloc.y+1*ws.height); Rimps[1].show();
      ImageWindow.setNextLocation(rloc.x+1*ws.width, rloc.y+0*ws.height); Rimps[2].show();
      for(ImagePlus rimp : Rimps)
         rimp.getWindow().setSize(ws.width,ws.height);
      for(ImagePlus rimp : Rimps)
         rimp.getWindow().getCanvas().fitToWindow();
      setupPlotting(imp,Rimps);
      }

   public static ImagePlus[] processStretch(ImagePlus imp) {

      ImagePlus mask = (new Duplicator()).run(imp,1,1,1,imp.getNSlices(),1,1);
      IJ.run(mask, "32-bit", "");
      IJ.run(mask,"Convert to Mask", "method=Huang background=Default calculate black");
      IJ.run(mask, "Fill Holes", "stack");
/*
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Dilate", "stack");
      IJ.run(mask, "Dilate", "stack");
      IJ.run(mask, "Dilate", "stack");
*/
      
      ImagePlus[] pimp = F_Project.compute(imp, new F_Project.Compute() {
           @Override
           public double[] compute(double[] x, double[] y) {
              if (x == null) return new double[3];
              return fitNonLinSE(x,arrNorm(y));
              }
           }, mask, new ImagePlus[0]);

      String title = imp.getTitle();

      pimp[0].setTitle("diffusion "+title);
      pimp[0].setDisplayRange(0E-3,3E-3);

      pimp[1].setTitle("shape "+title);
      pimp[1].setDisplayRange(0.5,1);

      pimp[2].setTitle("chi "+title);
      pimp[2].setDisplayRange(0.00,0.06);

      return pimp;
      }

//////////////////////////////////////////////////////////////////////////////////
   private void processIVIMStretchAndDisplay(ImagePlus imp) {
      Rimps = processIVIMStretch(imp);

      Dimension ws = imp.getWindow().getSize();
      Point rloc = new Point(loc.x - 4*ws.width, loc.y);
      ImageWindow.setNextLocation(rloc.x+0*ws.width, rloc.y+0*ws.height); Rimps[0].show();
      ImageWindow.setNextLocation(rloc.x+0*ws.width, rloc.y+1*ws.height); Rimps[1].show();
      ImageWindow.setNextLocation(rloc.x+1*ws.width, rloc.y+0*ws.height); Rimps[2].show();
      ImageWindow.setNextLocation(rloc.x+1*ws.width, rloc.y+1*ws.height); Rimps[3].show();
      ImageWindow.setNextLocation(rloc.x+1*ws.width, rloc.y+2*ws.height); Rimps[4].show();
      ImageWindow.setNextLocation(rloc.x+2*ws.width, rloc.y+0*ws.height); Rimps[5].show();
      for(ImagePlus rimp : Rimps)
         rimp.getWindow().setSize(ws.width,ws.height);
      for(ImagePlus rimp : Rimps)
         rimp.getWindow().getCanvas().fitToWindow();
      setupPlotting(imp,Rimps);
      }

   public static ImagePlus[] processIVIMStretch(ImagePlus imp) {

      ImagePlus mask = (new Duplicator()).run(imp,1,1,1,imp.getNSlices(),1,1);
      IJ.run(mask, "32-bit", "");
      IJ.run(mask,"Convert to Mask", "method=Huang background=Default calculate black");
      IJ.run(mask, "Fill Holes", "stack");
/*
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Erode", "stack");
      IJ.run(mask, "Dilate", "stack");
      IJ.run(mask, "Dilate", "stack");
      IJ.run(mask, "Dilate", "stack");
*/
      
      ImagePlus[] pimp = F_Project.compute(imp, new F_Project.Compute() {
           @Override
           public double[] compute(double[] x, double[] y) {
              if (x == null) return new double[6];

              return fitIVIMSE(x,arrNorm(y),NOISEFLOOR/arrMax(y));
              }
           }, mask, new ImagePlus[0]);

      String title = imp.getTitle();

      pimp[0].setTitle("fraction psuedo perfusion "+title);
      pimp[0].setDisplayRange(0,0.3);

      pimp[1].setTitle("pseudo perfusion "+title);
      pimp[1].setDisplayRange(3E-3,50E-3);

      pimp[2].setTitle("fraction diffusion "+title);
      pimp[2].setDisplayRange(0,1);

      pimp[3].setTitle("diffusion "+title);
      pimp[3].setDisplayRange(0E-3,3E-3);

      pimp[4].setTitle("shape "+title);
      pimp[4].setDisplayRange(0.5,1);

      pimp[5].setTitle("chi "+title);
      pimp[5].setDisplayRange(0.00,0.06);

      return pimp;
      }

//////////////////////////////////////////////////////////////////////////////////
   private void setupPlotting(ImagePlus imp, ImagePlus[] Rimps) {
      imp.getWindow().addWindowListener(this);
      imp.getWindow().getCanvas().addMouseListener(this);
      imp.getWindow().getCanvas().addMouseMotionListener(this);
      imp.getWindow().getCanvas().addKeyListener(this);

      engaged = true;

      for(ImagePlus rimp : Rimps)
         rimp.setZ(imp.getZ());
      for(ImagePlus rimp : Rimps)
         rimp.getWindow().addWindowListener(this);
      }

    boolean engaged = false;
    void disengage(ImageWindow iw) {
       if (!engaged) return;
        
       if (Rimps == null) {
          IJ.log("MultiModal_diffusion.disengage called with Rimps == null");
          return;
          }

       boolean oneOfUs = false;
       for(ImagePlus imp : Rimps)
          if (iw.getImagePlus() == imp)
             oneOfUs = true;
       if (!oneOfUs)
          return;

       if (dimp == null)
          IJ.showStatus("diffusion_MultiModal disengage DWI null");
       else {
          dimp.getWindow().getCanvas().removeMouseListener(this);
          dimp.getWindow().getCanvas().removeMouseMotionListener(this);
          dimp.getWindow().getCanvas().removeKeyListener(this);
          dimp.getWindow().removeWindowListener(this);

          engaged = false;
          IJ.showStatus("Multimodal Diffusion Disengaged");
          IJ.log("Multimodal Diffusion Disengaged");
          for(WindowListener wl : dimp.getWindow().getWindowListeners())
             IJ.log("WindowListeners: "+wl);
          }

       if (Rimps != null) {
          for(ImagePlus rimp : Rimps)
             rimp.getWindow().removeWindowListener(this);
          for(ImagePlus imp : Rimps) {
             if (imp.getWindow() == iw)
                continue;
             IJ.log("Closing: "+imp.getTitle());
             if (imp != null && imp.getWindow() != null) imp.getWindow().close();
             }
          }

       for(int i=0; i<pwv.size(); i++)
          pwv.get(i).close();
       }

//////////////////////////////////////////////////////////////////////////////////
   private void doPlotting(ImagePlus ROIimp) {
           if (model.equals("Multimodal"))   plotMultimodal(ROIimp);
      else if (model.equals("ADC"))          plotADC(ROIimp);
      else if (model.equals("IVIM"))         plotIVIM(ROIimp);
      else if (model.equals("Stretch"))      plotStretch(ROIimp);
      else if (model.equals("IVIM-Stretch")) plotIVIMStretch(ROIimp);
      else IJ.error("Unknown model '"+model+"'");
      }

   private void plotMultimodal(ImagePlus ROIimp) {
      try {

      int ix=0, iy=0;
      Roi roi = ROIimp.getRoi(); 
           if (roi == null) return;
      else if (roi.getType() == Roi.POINT) {
         //Point p = ((PointRoi)roi).getContainedPoints()[0];
         Rectangle p = roi.getBounds();
         plotVoxel(p.x,p.y,ROIimp.getZ(),ROIimp.getT());
         for(ImagePlus rimp : Rimps)
            if (rimp != null)
               rimp.setRoi((PointRoi)roi.clone()); //new PointRoi(p.x,p.y));
         return;
         } 
     else if (roi.getType() == Roi.POLYGON) {
         //Rectangle rec = roi.getBounds();
         //Roi BBRoi = new Roi(roi.getBounds());
         ColorProcessor mip = null;
         int w = 0;
         int h = 0;
         int[][] mipxy = new int[][]{{0,0},{0,1}, {1,0}, {2,0},{2,1},{2,2}, {3,0}, {4,1},{4,2},{4,2}};
         String[] labels = new String[]{"a) f_F","b) D_F","c) f_UI","d) f_H","e) D_H","f) \u03B1_H","g) f_R","i) \u03C7","j) r_xy","j) r_xy"};
         double[] multipliers = new double[]{1,1000,1,1,1000,1,1,1,1,1};
         Font font = new Font("Comic Sans MS",Font.BOLD,16);
         Rimps[0].setRoi(roi);
         for(int i=0; i<mipxy.length; i++) {
            Rimps[i].setRoi(roi);
            //mip.insert(Rimps[i].crop().getProcessor().convertToColorProcessor(),mipxy[i][0]*rec.width,mipxy[i][1]*rec.height);
            ColorProcessor cp = (ColorProcessor)(Colorizer_Overlay.doOverlayScaleLabelColorBar(new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
                                                                                               new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
                labels[i],Double.NaN,Double.NaN,multipliers[i],"<none>","100%x100%;30,30","None","Bottom-Center","Right","10%","%.1f",font,Color.white,Color.black
                ).getProcessor());
            if (mip == null) {
               w = cp.getWidth();
               h = cp.getHeight();
               mip = new ColorProcessor(5*w, 3*h);
               }
            mip.insert(cp,mipxy[i][0]*w,mipxy[i][1]*h);
            }
         //mip.insert(ROIimp.crop().getProcessor().convertToColorProcessor(),4*w,0*h);
         mip.insert((ColorProcessor)(Colorizer_Overlay.doOverlayScaleLabelColorBar(new ImagePlus("",ROIimp.getProcessor()),roi,new ImagePlus("",ROIimp.getProcessor()),roi,
                                        "h) b=0",Double.NaN,Double.NaN,1.0,null,"100%x100%;20,20","None","Bottom-Center",null,"","",font,Color.white,Color.black).getProcessor()),4*w,0*h);

         new ImagePlus("QuadModal_Montage",mip).show();
         return;
         }
     else {
         IJ.showStatus("Needs Point ROI or RECTANGLE");
         return;
         }
     //imp.setRoi(roi);

      }
      catch(Exception e) {
         StringWriter sw = new StringWriter();
         e.printStackTrace(new PrintWriter(sw));
         IJ.error("An error occurred.\n\n"+sw.toString());
         }
      }

   private void plotADC(ImagePlus ROIimp) {
      int ix=0, iy=0;
      Roi roi = ROIimp.getRoi(); 
           if (roi == null) return;
      else if (roi.getType() == Roi.POINT) {
         //Point p = ((PointRoi)roi).getContainedPoints()[0];
         Rectangle p = roi.getBounds();
         ix = p.x;
         iy = p.y;
         int iz = ROIimp.getZ();

         Plot plot = new Plot("DWI","b-value (s/mm^2)","ln(y_i/y_0)");
      
         plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
         plot.setXLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
         plot.setYLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
         plot.setLineWidth(2);
         plot.setSize(3*dx, 1*dy);
      
         if ( (pwv.size() < 1) || (pwv.get(0) == null)  || pwv.get(0).isClosed()) {
            Plot pl = new Plot("Diffusion ADC","","");
            pl.setSize(3*dx,1*dy);
            pwv.add(0,pl.show());
            pwDim = pwv.get(0).getSize();
            pwv.get(0).setLocation(loc.x,loc.y);
            }

         int NF = dimp.getNFrames();
         double[] x = new double[NF];
         double[] y = new double[NF];
         double y0 = 0;
         for(int f=0; f<NF; f++) {
            int si = dimp.getStackIndex(1,iz,f+1);
            x[f] = Double.parseDouble(dimp.getStack().getShortSliceLabel(si).split(" = |=| ",2)[1]);
            y[f] = (double)dimp.getStack().getProcessor(si).getPixelValue(ix,iy);
            }
         double[] Y = arrln(arrMul(y,1/y[0]));

         plot.addPoints(x,Y,Plot.X);

         double b0 = y[0];
         int[] i = new int[]{arrElementNear(x,1000),arrElementNear(x,2000),arrElementNear(x,3000)};
         double[] D = new double[]{-Y[i[0]]/x[i[0]],-Y[i[1]]/x[i[1]],-Y[i[2]]/x[i[2]]};

         plot.setColor(Color.red);
         plot.addLabel(0.6,0.1,d2s(D[0]*1000));
         plot.addPoints(new double[]{0,x[i[0]]},new double[]{0,Y[i[0]]},Plot.LINE);
         plot.setColor(Color.green);
         plot.addLabel(0.6,0.2,d2s(D[1]*1000));
         plot.addPoints(new double[]{0,x[i[1]]},new double[]{0,Y[i[1]]},Plot.LINE);
         plot.setColor(Color.blue);
         plot.addLabel(0.6,0.3,d2s(D[2]*1000));
         plot.addPoints(new double[]{0,x[i[2]]},new double[]{0,Y[i[2]]},Plot.LINE);

         plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
         plot.setColor(Color.black);
         plot.addLabel(0.03, 0.9, ""+ix+", "+iy+", "+iz+"\n");

         pwv.get(0).drawPlot(plot);

         for(ImagePlus rimp : Rimps)
            if (rimp != null)
               rimp.setRoi(roi);
         return;
         } 
     else if (roi.getType() == Roi.POLYGON) {
         //Rectangle rec = roi.getBounds();
         ColorProcessor mip = null;
         int w = 0;
         int h = 0;
         int[][] mipxy = new int[][]{{0,0},{0,1}, {1,0},{1,1}, {2,0},{2,1}};
         String[] labels = new String[]{"a) ADC_1000","b) chi_1000","c) ADC_2000","d) chi_2000","e) ADC_2800","f) chi_2800"};
         double[] multipliers = new double[]{1000,1,1000,1,1000,1};
         Font font = new Font("Arial",Font.BOLD,16);
         Rimps[0].setRoi(roi);
         for(int i=0; i<mipxy.length; i++) {
            Rimps[i].setRoi(roi);
            //mip.insert(Rimps[i].crop().getProcessor().convertToColorProcessor(),mipxy[i][0]*rec.width,mipxy[i][1]*rec.height);
//            ColorProcessor cp = (ColorProcessor)(Colorizer_Overlay.doOverlayLabelColorBar(null,null,new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
//                labels[i],Double.NaN,Double.NaN,multipliers[i],"<none>","Right","10%","%.2f",font,Color.white,Color.black
//                ).getProcessor());
            ColorProcessor cp = (ColorProcessor)(Colorizer_Overlay.doOverlayScaleLabelColorBar(new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
                                                                                               new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
                labels[i],Double.NaN,Double.NaN,multipliers[i],"<none>","100%x100%;20,20","None","Bottom-Center","Right","10%","%.1f",font,Color.white,Color.black
                ).getProcessor());
            if (mip == null) {
               w = cp.getWidth();
               h = cp.getHeight();
               mip = new ColorProcessor(3*w, 2*h);
               }
            mip.insert(cp,mipxy[i][0]*w,mipxy[i][1]*h);
            }

         new ImagePlus("ADC_Montage",mip).show();
         return;
         }
      }

   private void plotIVIM(ImagePlus ROIimp) {
      int ix=0, iy=0;
      Roi roi = ROIimp.getRoi(); 
           if (roi == null) return;
      else if (roi.getType() == Roi.POINT) {
         //Point p = ((PointRoi)roi).getContainedPoints()[0];
         Rectangle p = roi.getBounds();
         ix = p.x;
         iy = p.y;
         int iz = ROIimp.getZ();

         Plot plot = new Plot("IVIM DWI","b-value (s/mm^2)","ln(y_i/y_0)");
      
         plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
         plot.setXLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
         plot.setYLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
         plot.setLineWidth(2);
         plot.setSize(3*dx, 1*dy);
      
         if ( (pwv.size() < 1) || (pwv.get(0) == null)  || pwv.get(0).isClosed()) {
            Plot pl = new Plot("Diffusion IVIM","","");
            pl.setSize(3*dx,1*dy);
            pwv.add(0,pl.show());
            pwDim = pwv.get(0).getSize();
            pwv.get(0).setLocation(loc.x,loc.y);
            }

         int NF = dimp.getNFrames();
         double[] x = new double[NF];
         double[] y = new double[NF];
         double y0 = 0;
         for(int f=0; f<NF; f++) {
            int si = dimp.getStackIndex(1,iz,f+1);
            x[f] = Double.parseDouble(dimp.getStack().getShortSliceLabel(si).split(" = |=| ",2)[1]);
            y[f] = (double)dimp.getStack().getProcessor(si).getPixelValue(ix,iy);
            }
         double[] Y = arrln(arrMul(y,1/y[0]));

         plot.addPoints(x,Y,Plot.X);

         double[] params = fitIVIM(x,y);
         plot.addLabel(0.6,0.1,"ff\nDf\nfd\nDd");
         plot.addLabel(0.8,0.1," \num2/ms\n \num2/ms");
         plot.setColor(Color.red);
         plot.addLabel(0.7,0.1,d2s(params[0])+"\n"+d2s(params[1]*1000)+"\n"+d2s(params[2])+"\n"+d2s(params[3]*1000));
         double[] YY = arrln(arrSum(se(x,params[0],params[1],1),se(x,params[2],params[3],1)));
         plot.addPoints(x,YY,Plot.LINE);

         plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
         plot.setColor(Color.black);
         plot.addLabel(0.03, 0.9, ""+ix+", "+iy+", "+iz+"\n");
         pwv.get(0).drawPlot(plot);

         for(ImagePlus rimp : Rimps)
            if (rimp != null)
               rimp.setRoi(roi);
         return;
         } 
     else if (roi.getType() == Roi.POLYGON) {
         //Rectangle rec = roi.getBounds();
         ColorProcessor mip = null;
         int w = 0;
         int h = 0;
         int[][] mipxy = new int[][]{{0,0},{0,1}, {1,0},{1,1}, {2,0}};
         String[] labels = new String[]{"a) f_p","b) D*","c) f_d","d) D_d","e) chi"};
         double[] multipliers = new double[]{1,1000,1,1000,1};
         Font font = new Font("Arial",Font.BOLD,16);
         Rimps[0].setRoi(roi);
         for(int i=0; i<mipxy.length; i++) {
            Rimps[i].setRoi(roi);
            //mip.insert(Rimps[i].crop().getProcessor().convertToColorProcessor(),mipxy[i][0]**rec.width,mipxy[i][1]*rec.height);
//            ColorProcessor cp = (ColorProcessor)(Colorizer_Overlay.doOverlayLabelColorBar(null,null,new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
//                labels[i],Double.NaN,Double.NaN,multipliers[i],"Jet","Right","10%","%.2f",font,Color.white,Color.black
//                ).getProcessor());
            ColorProcessor cp = (ColorProcessor)(Colorizer_Overlay.doOverlayScaleLabelColorBar(new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
                                                                                               new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
                labels[i],Double.NaN,Double.NaN,multipliers[i],"<none>","100%x100%;20,20","None","Bottom-Center","Right","10%","%.1f",font,Color.white,Color.black
                ).getProcessor());
            if (mip == null) {
               w = cp.getWidth();
               h = cp.getHeight();
               mip = new ColorProcessor(3*w, 2*h);
               }
            mip.insert(cp,mipxy[i][0]*w,mipxy[i][1]*h);
            }

         new ImagePlus("IVIM_Montage",mip).show();
         return;
         }
      }


   private void plotStretch(ImagePlus ROIimp) {
      int ix=0, iy=0;
      Roi roi = ROIimp.getRoi(); 
           if (roi == null) return;
      else if (roi.getType() == Roi.POINT) {
         //Point p = ((PointRoi)roi).getContainedPoints()[0];
         Rectangle p = roi.getBounds();
         ix = p.x;
         iy = p.y;
         int iz = ROIimp.getZ();

         Plot plot = new Plot("Stretch DWI","b-value (s/mm^2)","ln(y_i/y_0)");
      
         plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
         plot.setXLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
         plot.setYLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
         plot.setLineWidth(2);
         plot.setSize(3*dx, 1*dy);
      
         if ( (pwv.size() < 1) || (pwv.get(0) == null)  || pwv.get(0).isClosed()) {
            Plot pl = new Plot("Diffusion Stretch","","");
            pl.setSize(3*dx,1*dy);
            pwv.add(0,pl.show());
            pwDim = pwv.get(0).getSize();
            pwv.get(0).setLocation(loc.x,loc.y);
            }

         int NF = dimp.getNFrames();
         double[] x = new double[NF];
         double[] y = new double[NF];
         double y0 = 0;
         for(int f=0; f<NF; f++) {
            int si = dimp.getStackIndex(1,iz,f+1);
            x[f] = Double.parseDouble(dimp.getStack().getShortSliceLabel(si).split(" = |=| ",2)[1]);
            y[f] = (double)dimp.getStack().getProcessor(si).getPixelValue(ix,iy);
            }
         double[] Y = arrln(arrMul(y,1/y[0]));

         plot.addPoints(x,Y,Plot.X);

         double[] params = fitNonLinSE(x,arrNorm(y));
         plot.addLabel(0.6,0.1,"Dd\nad");
         plot.addLabel(0.8,0.1," \num2/ms\n% ");
         plot.setColor(Color.red);
         plot.addLabel(0.7,0.1,d2s(params[0]*1000)+"\n"+d2s(params[1])+"\n"+d2s(params[2]*100));
         double[] YY = arrln(se(x,1,params[0],params[1]));
         plot.addPoints(x,YY,Plot.LINE);

         plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
         plot.setColor(Color.black);
         plot.addLabel(0.03, 0.9, ""+ix+", "+iy+", "+iz+"\n");

         pwv.get(0).drawPlot(plot);

         for(ImagePlus rimp : Rimps)
            if (rimp != null)
               rimp.setRoi(roi);
         return;
         } 
     else if (roi.getType() == Roi.POLYGON) {
         //Rectangle rec = roi.getBounds();
         ColorProcessor mip = null;
         int w = 0;
         int h = 0;
         int[][] mipxy = new int[][]{{0,0},{0,1}, {1,0}};
         String[] labels = new String[]{"a) D_s","b) alpha_s","c) chi"};
         double[] multipliers = new double[]{1000,1,1};
         Font font = new Font("Arial",Font.BOLD,16);
         Rimps[0].setRoi(roi);
         for(int i=0; i<mipxy.length; i++) {
            Rimps[i].setRoi(roi);
            //mip.insert(Rimps[i].crop().getProcessor().convertToColorProcessor(),mipxy[i][0]*rec.width,mipxy[i][1]*rec.height);
//            ColorProcessor cp = (ColorProcessor)(Colorizer_Overlay.doOverlayLabelColorBar(null,null,new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
//                labels[i],Double.NaN,Double.NaN,multipliers[i],null,"Right","10%","%.2f",font,Color.white,Color.black
//                ).getProcessor());
            ColorProcessor cp = (ColorProcessor)(Colorizer_Overlay.doOverlayScaleLabelColorBar(new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
                                                                                               new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
                labels[i],Double.NaN,Double.NaN,multipliers[i],"<none>","100%x100%;20,20","None","Bottom-Center","Right","10%","%.1f",font,Color.white,Color.black
                ).getProcessor());
            if (mip == null) {
               w = cp.getWidth();
               h = cp.getHeight();
               mip = new ColorProcessor(2*w, 2*h);
               }
            mip.insert(cp,mipxy[i][0]*w,mipxy[i][1]*h);
            }

         new ImagePlus("Stretch_Montage",mip).show();
         return;
         }
      }

   private void plotIVIMStretch(ImagePlus ROIimp) {
      int ix=0, iy=0;
      Roi roi = ROIimp.getRoi(); 
           if (roi == null) return;
      else if (roi.getType() == Roi.POINT) {
         //Point p = ((PointRoi)roi).getContainedPoints()[0];
         Rectangle p = roi.getBounds();
         ix = p.x;
         iy = p.y;
         int iz = ROIimp.getZ();

         Plot plot = new Plot("IVIM Stretch","b-value (s/mm^2)","ln(y_i/y_0)");
      
         plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
         plot.setXLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
         plot.setYLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
         plot.setLineWidth(2);
         plot.setSize(3*dx, 1*dy);
      
         if ( (pwv.size() < 1) || (pwv.get(0) == null)  || pwv.get(0).isClosed()) {
            Plot pl = new Plot("IVIM Stretch","","");
            pl.setSize(3*dx,1*dy);
            pwv.add(0,pl.show());
            pwDim = pwv.get(0).getSize();
            pwv.get(0).setLocation(loc.x,loc.y);
            }

         int NF = dimp.getNFrames();
         double[] x = new double[NF];
         double[] y = new double[NF];
         double y0 = 0;
         for(int f=0; f<NF; f++) {
            int si = dimp.getStackIndex(1,iz,f+1);
            x[f] = Double.parseDouble(dimp.getStack().getShortSliceLabel(si).split(" = |=| ",2)[1]);
            y[f] = (double)dimp.getStack().getProcessor(si).getPixelValue(ix,iy);
            }
         double[] Y = arrln(arrMul(y,1/y[0]));

         plot.addPoints(x,Y,Plot.X);

         double[] params = fitIVIMSE(x,arrNorm(y),NOISEFLOOR/arrMax(y));
         plot.addLabel(0.6,0.1,"fp\nD*\nfd\nDd\nad\nrms");
         plot.addLabel(0.8,0.1," \num2/ms\n \num2/ms\n \n %");
         plot.setColor(Color.red);
         plot.addLabel(0.7,0.1,d2s(params[0])+"\n"+d2s(params[1]*1000)
                         +"\n"+d2s(params[2])+"\n"+d2s(params[3]*1000)+"\n"+d2s(params[4])
                         +"\n"+d2s(params[5]*100));
         double[] YY = arrln(arrSum(se(x,params[0],params[1],1),se(x,params[2],params[3],params[4])));
         plot.addPoints(x,YY,Plot.LINE);

         plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
         plot.setColor(Color.black);
         plot.addLabel(0.03, 0.9, ""+ix+", "+iy+", "+iz+"\n");

         pwv.get(0).drawPlot(plot);

         for(ImagePlus rimp : Rimps)
            if (rimp != null)
               rimp.setRoi(roi);
         return;
         } 
     else if (roi.getType() == Roi.POLYGON) {
         //Rectangle rec = roi.getBounds();
         ColorProcessor mip = null;
         int w = 0;
         int h = 0;
         int[][] mipxy = new int[][]{{0,0},{0,1}, {1,0},{1,1},{1,2}, {2,0}};
         String[] labels = new String[]{"a) f_p","b) D*", "c) f_is", "d) D_is","e) alpha_is","c) chi_is"};
         double[] multipliers = new double[]{1,1000,1,1000,1,1};
         Font font = new Font("Arial",Font.BOLD,16);
         Rimps[0].setRoi(roi);
         for(int i=0; i<mipxy.length; i++) {
            Rimps[i].setRoi(roi);
            //mip.insert(Rimps[i].crop().getProcessor().convertToColorProcessor(),mipxy[i][0]*rec.width,mipxy[i][1]*rec.height);
//            ColorProcessor cp = (ColorProcessor)(Colorizer_Overlay.doOverlayLabelColorBar(null,null,new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
//                labels[i],Double.NaN,Double.NaN,multipliers[i],null,"Right","10%","%.2f",font,Color.white,Color.black
//                ).getProcessor());
            ColorProcessor cp = (ColorProcessor)(Colorizer_Overlay.doOverlayScaleLabelColorBar(new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
                                                                                               new ImagePlus("",Rimps[i].getProcessor()),Rimps[i].getRoi(),
                labels[i],Double.NaN,Double.NaN,multipliers[i],"<none>","100%x100%;20,20","None","Bottom-Center","Right","10%","%.1f",font,Color.white,Color.black
                ).getProcessor());
            if (mip == null) {
               w = cp.getWidth();
               h = cp.getHeight();
               mip = new ColorProcessor(3*w, 3*h);
               }
            mip.insert(cp,mipxy[i][0]*w,mipxy[i][1]*h);
            }

         new ImagePlus("IVIMStretch_Montage",mip).show();
         return;
         }
      }

/////////////////////////////
   public static String d2s(double x) {
      return String.format("%.2f",x);
      }

   public static double clamp(double min, double value, double max) {
      return value>max ? max : value<min ? min : value;
      }

/////////////////////////////

   public static double flowQuality(double[] b, double[] y, double noiseFloor) {
      double[][] by = arrSubSet(b,arrln(arrNorm(y)), new double[]{0, 150}, Math.log(noiseFloor));
      double[] mbrs = linreg(by[0], by[1]);
      return mbrs[3];
      }

   public static double[] nonlinearFit(double[] b, double[] y) {
      double b0v = arrMax(y);
      double noiseFloor = NOISEFLOOR;
      double nf = noiseFloor/b0v;
      double[] Y = arrNorm(y,noiseFloor);

      double ft = 1.0;

      double[] baseRange = new double[]{500,3000};
      double[] tailRange = new double[]{3000,b[b.length-1]};

      double Dr = 1E-7;
      double ar = 1.0;
      double fr = ft*fitMonoFraction2(b,Y,Dr,ar,baseRange,tailRange,nf);
      double frSave = fr;
      double[] y2 = arrDiff(Y,se(b,fr,Dr,1));
      double[] Y2 = arrNorm(y2);

      {
         double rms = rmse(arrSubSet(new double[][]{b,arrNorm(Y)},tailRange),arrSubSet(new double[][]{b,se(b,fr,Dr,1)},tailRange));

         double[][] diffby = arrDeNaN(arrSubSet(b,arrln(arrNorm(Y)),new double[]{4000,b[b.length-1]}, Math.log(nf)));
         double frFixed = Math.max(Math.exp(arrMean(diffby[1])) * Math.log(2),Math.exp(arrMin(diffby[1]))); 

         double[][] ls = arrDeNaN(linse(arrDeNoise(b,Y2,nf)));
         double[][] lsbase = arrSubSet(ls,arrln(baseRange));
         if (lsbase[1].length > 1 && Math.exp(ls[0][ls[0].length-1]) >= tailRange[0]) {
            double[] mbb = theilsen(lsbase);
            double bad = mbb[0];
            double bDd = Math.exp(mbb[1]/bad);
   
            if (Double.isNaN(bad) || Double.isNaN(bDd) || (bad < 0.5) || (bad > 1.0) || (bDd < 0.5e-3))
               fr = frFixed;
            else if ((fr < 0.06) && (rms > 10.0/100))
               fr = frFixed;
            }
         else if (ls[0].length > 0 && ls[0][ls[0].length-1] < Math.log(tailRange[1]))
               fr = frFixed;
         else if ((fr < 0.06 || fr > 0.65) && rms > 10.0/100)
               fr = frFixed;

         if (true) {
            double[][] lmnf = arrDeNaN(arrSubSet(b,arrln(Y),new double[]{1500,b[b.length-1]}));
            double[] mbb = theilsen(lmnf);
            double lmD = -mbb[0];
            double lmf = Math.exp(mbb[1]);
            if (lmD < 0.1e-3) {
               double[][] by = arrDeNaN(arrSubSet(b,arrln(Y),new double[]{4000,b[b.length-1]}));
               fr = Math.exp(arrMean(by[1])); //Math.max(Math.exp(arrMean(by[1])) * Math.log(2),Math.exp(arrMin(by[1])));
               }
         }

         y2 = arrDiff(Y,se(b,fr,Dr,1));
         Y2 = arrNorm(y2);
      }

      ft -= fr;

      double[] ffp = fitFluidFlow(b,Y2,nf);

      double ff = ffp[0]*ft;
      double Df = ffp[1];
      double af = 1.0;

      double fF = ffp[2]*ft;
      double DF = 3e-3;
      double aF = 1.0;

      double fd = ffp[3]*ft;
      double Dd = ffp[4];
      double ad = ffp[5];


      ft -= ff + fF;

      double[] y3 = arrDiff(y2,arrAdd(se(b,ff,Df,af),se(b,fF,DF,aF)));
      double[] Y3 = arrNorm(y3);

      double[] sep = fitLinSE(b,Y3,new double[]{500, 4000},nf);
      if (!((sep[0]*ft > 0.0) && (Double.isNaN(sep[1]) || (sep[1] > 3e-3) || (sep[2] < 0.5) || (sep[2] > 1.0)) ))  {
         fd = sep[0]*ft;
         Dd = sep[1];
         ad = Math.max(0.5,Math.min(sep[2],1));
         }
      ad = Math.max(0.5,Math.min(ad,1));
      Dd = clamp(0.1e-3,Dd,3e-3);
 
      if ((fd <0.05) && Double.isNaN(Dd)) {
         fd = 0;
         ft = ff + fF + fr;
         ff /= ft;
         fF /= ft;
         fr /= ft;
         }

      double[] yy = arrSum(se(b,ff,Df,af),se(b,fF,DF,aF),se(b,fd,Dd,ad),se(b,fr,Dr,ar));
      double rms = rmse(arrNorm(y),yy);

      double[] ssep = fitLinSE(b,Y2,new double[]{500,2500},noiseFloor);
      if (ssep[1] > 2.9e-3 || ssep[1] < 0.1e-3 || ssep[2] < 0.5 || ssep[2] > 1.0)
         ssep[3] = Double.POSITIVE_INFINITY;
      double ofrmse  = fitOnlyFluid(b,y);
      double frrmse  = fitOnlyFluid(b,Y2);

      switch(arrMinAt(new double[]{rms,ssep[3],ofrmse,frrmse})) {
         case 0: break;
//         case 1: ff=0.0; Df=Double.NaN; fF=0.0;    fd=1.0-fr; Dd=ssep[0];    ad=ssep[1];           rms=ssep[3]; break;
         case 2: ff=0.0; Df=Double.NaN; fF=1.0;    fd=0.0; Dd=Double.NaN; ad=Double.NaN; fr=0.0; rms=ofrmse;  break; 
         case 3: ff=0.0; Df=Double.NaN; fF=1.0-fr; fd=0.0; Dd=Double.NaN; ad=Double.NaN;         rms=frrmse;  break; 
         }
      if (fd == 0.0)
         ad = Double.NaN;

      noiseFloor = b0v*sep[3]/b.length; // when rmse is fixed bo not divide by b.length

      double flowbad = flowQuality(b,y,nf);

      return new double[]{ff,Df,fF,fd,Dd,ad,fr,rms,noiseFloor,flowbad}; //b0p};
      }


   Vector<PlotWindow> pwv = new Vector<PlotWindow>();
   Dimension pwDim = new Dimension();
   private void plotVoxel(int ix, int iy, int iz, int it) {
      PlotWindow.noGridLines = false; // draw grid lines
      Plot plot = new Plot("DWI","b-value (s/mm^2)","ln(y_i/y_0)");
      
      plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
      plot.setXLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
      plot.setYLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
      plot.setLineWidth(2);
      plot.setWindowSize(3*dx, 1*dy);
      
      if ( (pwv.size() < 1) || (pwv.get(0) == null)  || pwv.get(0).isClosed()) {
         Plot p = new Plot("Diffusion MulitModal","","");
         p.setWindowSize(3*dx, 1*dy);
         pwv.add(0,p.show());
         //pwv.get(0).setSize(3*dx,1*dy);
         pwDim = pwv.get(0).getSize(); pwDim.width -= 9; pwDim.height += 8; // OS adornments??
         pwv.get(0).setLocation(loc.x,loc.y);
         pwv.get(0).windowActivated(null);
         }

      try {
         int NF = dimp.getNFrames();
         double[] b = new double[NF];
         double[] y = new double[NF];
         double y0 = 0;
         for(int f=0; f<NF; f++) {
            int si = dimp.getStackIndex(1,iz,f+1);
            b[f] = Double.parseDouble(dimp.getStack().getShortSliceLabel(si).split(" = |=| ",2)[1]);
            y[f] = (double)dimp.getStack().getProcessor(si).getf(ix,iy); //getPixelValue(ix,iy);
            }
         
         double[] ep = new double[]{ Rimps[0].getStack().getProcessor(iz).getPixelValue(ix,iy),
                                     Rimps[1].getStack().getProcessor(iz).getPixelValue(ix,iy),
                                     Rimps[2].getStack().getProcessor(iz).getPixelValue(ix,iy),
                                     Rimps[3].getStack().getProcessor(iz).getPixelValue(ix,iy),
                                     Rimps[4].getStack().getProcessor(iz).getPixelValue(ix,iy),
                                     Rimps[5].getStack().getProcessor(iz).getPixelValue(ix,iy),
                                     Rimps[6].getStack().getProcessor(iz).getPixelValue(ix,iy),
                                     Rimps[7].getStack().getProcessor(iz).getPixelValue(ix,iy) };


         plot.setColor(Color.black);
         plot.setLineWidth(3);
         plot.addPoints(b,arrln(arrNorm(y)),Plot.X);

         double[] YY = arrln(arrAdd(se(b,ep[0],ep[1],1.0),se(b,ep[2],3e-3,1),se(b,ep[3],ep[4],ep[5]),se(b,ep[6],1E-7,1.0)));
         plot.setColor(Color.red);
         plot.setLineWidth(2);
         plot.addPoints(b,YY,Plot.LINE);


         // fix b=0

         double nf = Math.log(NOISEFLOOR/arrMax(y));  
         plot.setColor(Color.black);    
         plot.setLineWidth(3);
         plot.addPoints(new double[]{b[0],b[b.length-1]},new double[]{nf,nf},Plot.LINE);

         y = arrNorm(y,NOISEFLOOR);

         double flowbad = flowQuality(b,y,Math.exp(nf));
         plot.setFont(new Font("Comic Sans MS", Font.BOLD, 20));
         plot.addLabel(0.0,0.0,String.format("Low b bad = %4.2f", flowbad));
         plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
         if (false) {
            double[][] by = arrSubSet(b,arrln(y), new double[]{0, 150}, nf);
            double[] mbrs = linreg(by[0], by[1]);
            //flowbad = 1.0/( (mbrs[0]<0.0?0.1:1) * mbrs[2] );
            flowbad = mbrs[3];
            plot.addLabel(0.075,0.075,String.format("-(%5.2f \u00b5m\u00b2/ms)b + ln(%5.2f) (%5.2f)\n          %5.2f @%g", 
                                                    -mbrs[0]*1e3, Math.exp(mbrs[1]), mbrs[2], flowbad, b[arrMaxAt(y)]));
            }

         double[][] by = arrDeNaN(b,y);
         plot.setLineWidth(2);
         plot.addPoints(by[0],arrln(by[1]),Plot.X);

         double yhr = Double.NaN;
         double yhrm = Double.NaN;
         {
           double[][] diffby = arrDeNaN(arrSubSet(b,arrln(y),new double[]{4000,b[b.length-1]}, nf));
           yhr = Math.log(Math.max(Math.exp(arrMean(diffby[1])) * Math.log(2),Math.exp(arrMin(diffby[1])))); //center(diffby[1]);
           yhrm = arrMean(diffby[1]);
           plot.setColor(Color.magenta);
           plot.addPoints(new double[]{b[0],b[b.length-1]}, new double[]{yhr, yhr}, Plot.LINE);
           plot.setFont(new Font("Comic Sans MS", Font.BOLD, 16));
           plot.setJustification(Plot.RIGHT);
           plot.addText("f_HR "+String.format("%5.2f\n%5.2f",Math.exp(yhr),Math.exp(arrMean(diffby[1]))), b[b.length-1], yhr);
           plot.setJustification(Plot.LEFT);
         }

         double ft = 1.0;

         double[] Y = y; //arrNorm(arrDiff(y,se(b,ff,Df,1F)));

         double[] baseRange = new double[]{500,3000};
         double[] tailRange = new double[]{3000,b[b.length-1]};
         double Dr = 1E-7;
         double ar = 1.0;
         double fr = ft*fitMonoFraction2(b,Y,Dr,ar,baseRange,tailRange,Math.exp(nf));
         if (Double.isNaN(fr)) fr = 0.0;

         plotResidual(1,loc.x+pwDim.width,loc.y+0*pwDim.height,b,Y,baseRange,tailRange, fr, Dr, ar, Math.exp(nf));

         double[] y2 = arrDiff(Y,se(b,fr,Dr,1));
         double[] Y2 = arrNorm(y2);

         if (true) {
            double[][] ls = arrDeNaN(linse(arrDeNoise(b,Y2,Math.exp(nf))));
            double[][] lsbase = arrSubSet(ls,arrln(baseRange));

            double rms = Math.sqrt(arrMeanSquareDiff(arrSubSet(new double[][]{b,y},tailRange),arrSubSet(new double[][]{b,se(b,fr,Dr,1)},tailRange)));
//IJ.log("RMS = "+(rms*100)+"  fr = "+fr);
            double[][] diffby = arrDeNaN(arrSubSet(b,arrln(arrNorm(y)),new double[]{4000,b[b.length-1]}, nf));
            double frFixed = Math.max(Math.exp(arrMean(diffby[1])) * Math.log(2),Math.exp(arrMin(diffby[1]))); 

            if (lsbase[1].length > 1 && Math.exp(ls[0][ls[0].length-1]) >= tailRange[0]) {
               double[] mbb = theilsen(lsbase);
               double bad = mbb[0];
               double bDd = Math.exp(mbb[1]/bad);
   
               plot.addText(String.format("alpha=%5.2f D=%5.2f  rms=%5.2f",bad,bDd*1e3, rms*100), b[0], yhr);
               if (Double.isNaN(bad) || Double.isNaN(bDd) || (bad < 0.5) || (bad > 1.0) || (bDd < 0.5e-3)) {
                  fr = frFixed;
                  plot.addText(" \nOverride ", b[0], yhr);
                  }
               else if ((fr < 0.06 || fr > 0.65) && (rms > 10.0/100)) {
                  fr = frFixed;
                  plot.addText(" \nOverride fr equals zero or too big and big rms", b[0], yhr);
                  }
               }
            else if (ls[0].length > 0 && ls[0][ls[0].length-1] < Math.log(tailRange[0])) {
               fr = frFixed;
               plot.addText(" \nOverRide No Tail Range", b[0], yhr);
               }
            else if ((fr < 0.06 || fr > 0.65) && rms > 10.0/100) {
               fr = frFixed;
               plot.addText(" \nOverRide fr equals zero or too big and big rms "+(rms*100), b[0], yhr);
               }

            {
               double[][] lmnf = arrDeNaN(arrSubSet(b,arrln(y),new double[]{1500,b[b.length-1]}));
               double[] mbb = theilsen(lmnf);
               double lmD = -mbb[0];
               double lmf = Math.exp(mbb[1]);
               if (true) {
                  plot.setFont(new Font("Monospaced", Font.BOLD, 18));
                  String str = "";
                  if (lmD < 0.1e-3) {
                     fr = Math.exp(yhrm); //Math.exp(yhr);
                     str = "Override ";
                     }
                  plot.addText(str+String.format("D=%5.2f f=%5.2f", lmD*1e3, lmf),b[14],yhr);
                  }
            }
            y2 = arrDiff(Y,se(b,fr,Dr,1));
            Y2 = arrNorm(y2);
         }

         ft -= fr;

         double[] ffp = plotFluid(2,loc.x+0*pwDim.width,loc.y+1*pwDim.height,b,Y2, Math.exp(nf));

         double ff = ffp[0]*ft;
         double Df = ffp[1];
         double af = 1.0;

         double fF = ffp[2]*ft;
         double DF = ffp[3];
         double aF = 1.0;

         double fd = ffp[4]*ft;
         double Dd = ffp[5];
         double ad = 1.0;

         ft -= ff + fF;

         double[] y3 = arrDiff(y2,arrAdd(se(b,ff,Df,af),se(b,fF,DF,aF)));
         double[] Y3 = arrNorm(y3);

         double[] yB = arrDiff(y2,se(b,fF,DF,aF));
         double[] YB = arrNorm(y3);

         double[] sep = plotLinSE(3,loc.x+1*pwDim.width,loc.y+1*pwDim.height,b,YB, Math.exp(nf));

         if (!((sep[0]*ft > 0.0) && (Double.isNaN(sep[1]) || (sep[1] > 3e-3) || (sep[2] < 0.5) || (sep[2] > 1.0)) ))  {
            fd = sep[0]*ft;
            Dd = sep[1];
            ad = Math.max(0.5,Math.min(sep[2],1));
            }
         if (fd == 0D) fd = 0D; // -0 * something is not 0
         ad = Math.max(0.5,Math.min(ad,1));
         //if (Double.doubleToIntBits(fd) != 0) IJ.log("UGH");

         double noiseFloor = Math.log(sep[3]);
         if (noiseFloor > nf) {
            plot.setColor(Color.red);
            plot.setLineWidth(3);
            plot.addPoints(new double[]{b[0],b[b.length-1]},new double[]{noiseFloor,noiseFloor},Plot.LINE);
            }


         double[] yyy = arrSum(se(b,ff,Df,af),se(b,fF,DF,aF),se(b,fd,Dd,ad),se(b,fr,Dr,ar));
         plot.setColor(Color.blue);
         plot.addPoints(b,arrln(arrNorm(yyy)),Plot.LINE);
         double rms = rmse(y,arrNorm(yyy));

         plot.setFont(new Font("Monospaced", Font.BOLD, 18));
         plot.setColor(Color.black);
         plot.setJustification(Plot.RIGHT);
         double lh = 0.075;
         plot.addLabel(0.999, 0.075, "     SI"
                                    +"     ff"
                                    +"     Df"
                                    +"     fF"
                                    +"     fd"
                                    +"     Dd"
                                    +"  alpha"
                                    +"     fr" 
                                    +" rms(%)");


         plot.setColor(Color.blue);
         plot.addLabel(0.999, 0.075+1*lh,  "      "
                                 +"  "+String.format("%5.2f",ff)
                                 +"  "+String.format("%5.2f",Df*1e3F)
                                 +"  "+String.format("%5.2f",fF)
                                 +"  "+String.format("%5.2f",fd)
                                 +"  "+String.format("%5.2f",Dd*1e3F)
                                 +"  "+String.format("%5.2f",ad)
                                 +"  "+String.format("%5.2f",fr) 
                                 +"  "+String.format("%5.2f",rms*100));


         plot.setColor(Color.red);
         plot.addLabel(0.999, 0.075+2*lh, String.format("%7.0f", dimp.getStack().getProcessor(dimp.getStackIndex(1,iz,it)).getPixelValue(ix,iy))
                           +"  "+String.format("%5.2f",ep[0])
                           +"  "+String.format("%5.2f",ep[1]*1e3F)
                           +"  "+String.format("%5.2f",ep[2])
                           +"  "+String.format("%5.2f",ep[3])
                           +"  "+String.format("%5.2f",ep[4]*1e3F)
                           +"  "+String.format("%5.2f",ep[5])
                           +"  "+String.format("%5.2f",ep[6]) 
                           +"  "+String.format("%5.2f",ep[7]*100));

         plot.setColor(Color.black);
         plot.addLabel(0.999, 0.075+3*lh, "  a.u."
                                    +"       "
                                    +" \u00b5m\u00b2/ms"
                                    +"       "
                                    +"       "
                                    +" \u00b5m\u00b2/ms"
                                    +"       "
                                    +"       " 
                                    +"      %");

         plot.setFont(new Font("Comic Sans MS", Font.BOLD, 20));
 
         plot.setJustification(Plot.RIGHT);
         plot.setColor(Color.black);
         plot.addLabel(0.25, 0.95, "\n \n \n \n "+ix+", "+iy+", "+iz+" NoiseFloor "+String.format("%3.1f",nf));

         {
            double yMin = Math.log(arrMin(new double[]{arrMin(y),Math.exp(YY[YY.length-1]),yyy[yyy.length-1]}));
            plot.setLimits(Double.NaN, Double.NaN, yMin*1.05, 0.0);
         }
         pwv.get(0).drawPlot(plot);
         }
      catch(Exception e) {
         StringWriter sw = new StringWriter();
         e.printStackTrace(new PrintWriter(sw));
         IJ.error("An error occurred.\n\n"+sw.toString());
         }

      }

//////////
   public static double[] arrSum(double[] ... arrs) {
      if (arrs.length < 1)
         return null;
      double[] res = new double[arrs[0].length];
      for(double[] a : arrs)
         for(int i=0; i<res.length; i++)
            res[i] += a[i];
      return res;
      }

//////////
   private double plotResidual(int n, int sx, int sy, double[] b, double[] y, double[] baseRange, double[] tailRange, double fr, double Dr, double alpha, double nf) {
      PlotWindow.noGridLines = false; // draw grid lines
      Plot plot = new Plot("DWI","ln b-value (s/mm^2)","ln(-ln(y_i/y_0))");
      
      plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
      plot.setXLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
      plot.setYLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
      plot.setLineWidth(2);
      plot.setWindowSize(3*dx, 1*dy);
      
      if ( (pwv.size() < (n+1)) || (pwv.get(n) == null)  || pwv.get(n).isClosed()) {
         Plot p = new Plot("Highly Restricted Diffusion","","");
         p.setWindowSize(3*dx,1*dy);
         pwv.add(n,p.show());
         //pwv.get(n).setSize(3*dx,1*dy);
         //pwDim = pwv.get(n).getSize();
         pwv.get(n).setLocation(sx,sy);
         //pwv.get(n).windowActivated(null);
         }

      try {
         double[][] ls = arrDeNaN(linse(b,y));
         plot.setLineWidth(1);
         plot.addPoints(ls[0],ls[1],Plot.X);

         ls = arrDeNaN(linse(arrDeNoise(b,y,Math.log(nf))));
         plot.setLineWidth(2);
         plot.addPoints(ls[0],ls[1],Plot.X);

         plot.setFont(new Font("Comic Sans MS", Font.BOLD, 14));
         plot.setColor(Color.black);
         plot.setJustification(Plot.RIGHT);
         plot.addLabel(0.6, 0.825, "Dd\nalpha");


         if (ls[0][ls[0].length-1] < Math.log(baseRange[0])) {
            plot.addLabel(0.5,0.2,"No Base Values");
            }
         else {
            double[][] lsbase = arrSubSet(ls,arrln(baseRange));
            plot.setColor(Color.blue);
            plot.setLineWidth(3);
            plot.addPoints(lsbase[0],lsbase[1],Plot.CIRCLE);
            double[] mbb = theilsen(lsbase);
            plot.setLineWidth(1);
            plot.addPoints(ls[0],makeLine(ls[0],mbb),Plot.LINE);
            double bad = mbb[0];
            double bDd = Math.exp(mbb[1]/bad);

            plot.setColor(Color.blue);
            plot.addLabel(0.675, 0.825, String.format("%5.2f",bDd*1e3)
                                +"\n"+String.format("%5.2f",bad));
            }

         if (ls[0][ls[0].length-1] < Math.log(tailRange[0])) {
            plot.addLabel(0.5,0.2,"No Tail Values");
            }
         else {
            double[][] lstail = arrSubSet(ls,arrln(tailRange));
            plot.setColor(Color.green);
            plot.setLineWidth(3);
            plot.addPoints(lstail[0],lstail[1],Plot.CIRCLE);
            double[] mbt = theilsen(lstail);
            plot.setLineWidth(1);
            plot.addPoints(ls[0],makeLine(ls[0],mbt),Plot.LINE);
            double tad = mbt[0];
            double tDd = Math.exp(mbt[1]/tad);

            plot.setColor(Color.green);
            plot.addLabel(0.75, 0.825,  String.format("%5.2f",tDd*1e3)
                                +"\n"+String.format("%5.2f",tad));
            }

         double rms = arrLinearTest(b, y, fr, Dr, alpha, baseRange, tailRange, Math.log(nf));
         plot.setColor(Color.red);
         plot.setJustification(Plot.RIGHT);
         plot.addLabel(0.7, 0.7, "rms\nnoise floor");
         plot.setJustification(Plot.LEFT);
         plot.addLabel(0.725, 0.7, String.format("%7.5f\n%7.5f",100*rms,Math.log(nf)));

         double[][] ls2 = arrDeNaN(linse(b,arrNorm(arrDiff(y,se(b,fr,Dr,alpha)))));
         plot.setColor(Color.red);
         plot.setLineWidth(1);
         plot.addPoints(ls2[0],ls2[1],Plot.CIRCLE);
         ls2 = arrDeNaN(linse(arrDeNoise(b,arrNorm(arrDiff(y,se(b,fr,Dr,alpha))),Math.log(nf/(1-fr)))));
         plot.setLineWidth(2);
         plot.addPoints(ls2[0],ls2[1],Plot.CIRCLE);
         {
            //double[][] by = arrDeNoise(b,arrNorm(arrDiff(y,se(b,fr,Dr,alpha))),Math.log(Math.exp(-3.5)/(1-fr)));
            //double[] mb  = theilsen(arrSubSet(arrDeNaN(linse(by)),rl));
            double[] mbb = theilsen(arrSubSet(ls2,arrln(baseRange)));
            plot.setLineWidth(1);
            plot.addPoints(ls2[0],makeLine(ls2[0],mbb),Plot.LINE);
            double ad = mbb[0];
            double Dd = Math.exp(mbb[1]/ad);

            plot.setColor(Color.red);
            plot.setJustification(Plot.RIGHT);
            plot.addLabel(0.85, 0.825,  String.format("%5.2f",Dd*1e3)
                                +"\n"+String.format("%5.2f",ad));
         }

         plot.setColor(Color.black);
         plot.setJustification(Plot.LEFT);
         plot.addLabel(0.875, 0.8, "\u00b5m\u00b2/ms\n "+String.format("%5.2f",fr));
         plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));

         pwv.get(n).drawPlot(plot);
         plot.setLimitsToFit(true);
         return rms;
         }
      catch(Exception e) {
         StringWriter sw = new StringWriter();
         e.printStackTrace(new PrintWriter(sw));
         IJ.error("An error occurred.\n\n"+sw.toString());
         }

      pwv.get(n).drawPlot(plot);
      plot.setLimitsToFit(true);
      return -1.0;
      }
      
   public static double[] fitSubLinearSE(double[] b, double[] y, double[] range, double noiseFloor) {
      double[][] by = arrDeNaN(arrSubSet(b,y,range,noiseFloor));
      if (by[0].length < 2) return null;
      double bo = by[0][0];
      double yo = by[1][0];
      double[][] by2 = new double[2][];
      by2[0] = arrSub(by[0],bo);
      by2[1] = arrDiv(by[1],yo);
      double[][] lss = arrDeNaN(linse(by2));
      double[] mb = theilsen(lss);
      double ad = mb[0];
      double Dd = Math.exp(mb[1]/ad);
      double fd = yo/Math.exp(-Math.pow(bo*Dd,ad));
//IJ.log("fd");
//      for(int i=0; i<by[0].length; i++)
//         IJ.log(""+by[0][i]+"  "+(by[1][i]/Math.exp(-Math.pow(by[0][i]*Dd,ad))));
      return new double[]{fd,Dd,ad};
      }

   public static double[] fitFluidFlow(double[] b, double[] y, double noiseFloor) {

      double DF = 3E-3;
      double aF = 1.0;

      double nd_fF = fitFluid(b,y,DF,noiseFloor);
      double nd_ff = 1-nd_fF;
      double nd_Df = Double.NaN;
      if (nd_fF < 1) nd_Df = fitFlow(b,arrNorm(arrDiff(y,se(b,nd_fF,DF,1))));
                else nd_Df = Double.NaN;

      double[][] by = arrDeNaN(b,y);
      double nd_rmse = Math.sqrt(arrMeanSquareDiff(by[1],arrAdd(se(by[0],nd_fF,DF,aF),se(by[0],nd_ff,nd_Df,1))));
//IJ.log("nd_fF="+nd_fF+"   nd_Df="+nd_Df+"   "+nd_rmse);


      double nFd_ff = 1;
      double nFd_Df = fitFlow(b,y);
      double nFd_rmse = Math.sqrt(arrMeanSquareDiff(by[1],se(by[0],nFd_ff,nFd_Df,1)));
      if (nFd_Df < 4e-3)
         nFd_rmse *= 2;
//IJ.log("nFd_Df="+nFd_Df+"   "+nFd_rmse);
      

      double slse_fd = Double.NaN;
      double slse_Dd = Double.NaN;
      double slse_ad = Double.NaN;
      double slse_fF = Double.NaN;
      double slse_ff = Double.NaN;
      double slse_Df = Double.NaN;
      double slse_af = Double.NaN;
      double slse_rmse = Double.NaN;
      double[] fdDdad = fitSubLinearSE(b,y,new double[]{1200,b[b.length-1]}, noiseFloor);
      if (fdDdad != null) {
//IJ.log("fd = "+fdDdad[0]+"  Dd = "+fdDdad[1]+"  ad = "+fdDdad[2]);
         slse_fd = fdDdad[0];
         slse_Dd = fdDdad[1];
         slse_ad = fdDdad[2];
         double[] slse_y2 = arrNorm(arrDiff(y,se(b,slse_fd,slse_Dd,1)));
         slse_fF = (1-slse_fd)*fitFluid(b,slse_y2,DF,noiseFloor);
//IJ.log("slse_fd="+slse_fd+"  slse_fF="+slse_fF);

         double[] slse_y3 = arrNorm(arrDiff(y,arrAdd(se(b,slse_fd,slse_Dd,1),se(b,slse_fF,DF,1))));

         slse_ff = 1 - slse_fF - slse_fd;
         slse_Df = Double.NaN;
         slse_af = 1;
         if (slse_ff < 0.01) slse_ff = 0;
         else slse_Df = fitFlow(b,slse_y3);

         if ((slse_ff > 0) && Double.isNaN(slse_Df))
            slse_Df = 1;

         slse_rmse = Math.sqrt(arrMeanSquareDiff(by[1],arrAdd(se(by[0],slse_fd,slse_Dd,slse_ad),se(by[0],slse_fF,DF,aF),se(by[0],slse_ff,slse_Df,slse_af))));
//IJ.log("slse_rmse="+slse_rmse);
         }

      double le_fd = 0;
      double le_Dd = Double.NaN;
      double le_ad = 1.0;
      double[][] diffby = arrDeNaN(arrSubSet(b,y,new double[]{1200, 5000}, noiseFloor));

      if ((diffby[0].length > 1) && (arrMax(diffby[0]) > 1500)) {
         double[] mb = theilsen(diffby[0],arrln(diffby[1]));
         le_fd = Math.exp(mb[1]);
         le_Dd = -mb[0];
//IJ.log("le_fd="+le_fd+"   le_Dd="+le_Dd);
         if ((le_Dd > 3e-3) || (le_Dd < 0) || Double.isNaN(le_fd)) {
            le_fd = 0;
            le_Dd = Double.NaN;
            }
         else if (le_fd > 1) {
            le_fd = 1;
            double[][] diff2by = arrDeNaN(arrSubSet(b,y,new double[]{300, 2500}, noiseFloor));
            if (diff2by[0].length > 0) 
               le_Dd = -centerM(diff2by[0],arrln(diff2by[1]));
//IJ.log("le_f d = "+le_fd+" le_D d = "+le_Dd);
            }
         }
////////////////////////////////////////////////
      else {
         double[][] diff2by = arrDeNaN(arrSubSet(b,y,new double[]{300, 2500}, noiseFloor));
         double[] mb = theilsen(diff2by[0],arrln(diff2by[1]));
         le_fd = Math.exp(mb[1]);
         le_Dd = -mb[0];
         if ((le_Dd > 2.75e-3) || (le_Dd < 0) || Double.isNaN(le_fd)) {
            le_fd = 0;
            le_Dd = Double.NaN;
            }
         }
////////////////////////////////////////////////

      double[] le_y2 = arrNorm(arrDiff(y,se(b,le_fd,le_Dd,1)));
      double le_fF = (1-le_fd)*fitFluid(b,le_y2,DF,noiseFloor);
//IJ.log("le_fd="+le_fd+"  le_fF="+le_fF);

      double[] le_y3 = arrNorm(arrDiff(y,arrAdd(se(b,le_fd,le_Dd,1),se(b,le_fF,DF,aF))));

      double le_ff = 1 - le_fF - le_fd;
      double le_Df = Double.NaN;
      double le_af = 1;
      if (le_ff < 0.01) le_ff = 0;
      else le_Df = fitFlow(b,le_y3);

      if ((le_ff > 0) && Double.isNaN(le_Df))
         le_Df = 1;

      double le_rmse = Math.sqrt(arrMeanSquareDiff(by[1],arrAdd(se(by[0],le_fd,le_Dd,le_ad),se(by[0],le_fF,DF,aF),se(by[0],le_ff,le_Df,le_af))));
//IJ.log("le_rmse="+le_rmse);


      double ff = 0;
      double Df = Double.NaN;
      double fF = 0;
      double fd = 0;
      double Dd = Double.NaN;
      double ad = Double.NaN;
      switch(arrMinAt(new double[]{nd_rmse,nFd_rmse,slse_rmse,le_rmse})) {
         case 0: ff=nd_ff;   Df=nd_Df;   fF=nd_fF;   fd=0;       Dd=Double.NaN; ad=Double.NaN; break;
         case 1: ff=nFd_ff;  Df=nFd_Df;  fF=0;       fd=0;       Dd=Double.NaN; ad=Double.NaN; break;
         case 2: ff=slse_ff; Df=slse_Df; fF=slse_fF; fd=slse_fd; Dd=slse_Dd;    ad=slse_ad;    break;
         case 3: ff=le_ff;   Df=le_Df;   fF=le_fF;   fd=le_fd;   Dd=le_Dd;      ad=1;          break;
         }
      if (fd == 0)
         ad = Double.NaN;
 
//IJ.log("return "+ff+"  "+Df+"  "+fF+"  "+fd+"  "+Dd);
      return new double[]{ff,Df,fF,fd,Dd,ad};
      }

   public static double fitFlow(double[] b, double[] y) {
      double[][] flowby = arrSubSet(b,y,new double[]{0,150});
      double sum = 0;
      int N = 0;
      for(int i=0; i<flowby[0].length; i++) {
         if (flowby[0][i] == 0)
             continue;
         double D = -Math.log(flowby[1][i])/flowby[0][i];
         if (!Double.isNaN(D) && (D > 3e-3) && (D < 100e-3)) {
            sum += D;
            N++;
            }
         }
      return (N>0) ? sum/N : 1;
      }

   public static double fitFluid(double[] b, double[] y, double DF, double noiseFloor) {
     double[][] fluidby = arrDeNaN(arrSubSet(b,y,new double[]{300, 1500},noiseFloor));
     return (fluidby[0].length > 0) ? Math.min(Math.exp(centerB(fluidby[0],arrln(fluidby[1]),-DF)),1) : 0;
     }

   private double[] plotFluid(int n, int sx, int sy, double[] b, double[] y, double noiseFloor) {
      PlotWindow.noGridLines = false; // draw grid lines
      Plot plot = new Plot("Fluid","b-value (ms/\u03bcm\u00b2)","ln(y_i/y_0)");
      
      plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
      plot.setXLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
      plot.setYLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
      plot.setLineWidth(2);
      plot.setWindowSize(3*dx, 1*dy);
      
      if ( (pwv.size() < (n+1)) || (pwv.get(n) == null)  || pwv.get(n).isClosed()) {
         Plot p = new Plot("Flow and Fluid Diffusion","","");
         p.setWindowSize(3*dx,1*dy);
         pwv.add(n,p.show());
         //pwv.get(n).setSize(3*dx,1*dy);
         //pwDim = pwv.get(n).getSize();
         pwv.get(n).setLocation(sx,sy);
         pwv.get(n).windowActivated(null);
         }

      try {
         plot.setFont(new Font("Monospaced", Font.BOLD, 18));
         double lh = 0.075;

         double DF = 3E-3;
         double aF = 1.0;

         plot.setColor(Color.black);
         plot.setLineWidth(2);
         plot.addPoints(b,arrln(y),Plot.X);

         plot.addPoints(new double[]{b[0],b[b.length-1]},arrln(new double[]{noiseFloor,noiseFloor}),Plot.LINE);

         double[][] by = arrDeNaN(b,y);

         plot.setColor(Color.black);
         plot.addLabel(0.01,0.71+0*lh,String.format("                 ff      Df    fF   fd    Dd    ad  rmse%%"));
         plot.addLabel(0.01,0.71+5*lh,String.format("                     um2/ms            um2/ms             "));
         

         double nFd_ff = 1;
         double nFd_Df = fitFlow(b,y);
         double nFd_rmse = rmse(by[1],se(by[0],nFd_ff,nFd_Df,1));

         plot.setColor(Color.green);
         plot.setLineWidth(1);
         plot.addPoints(b,arrln(se(b,nFd_ff,nFd_Df,1)),Plot.LINE);
         plot.addLabel(0.01,0.71+1*lh,String.format("Flow          %5.2f %7.2f                         %5.2f",nFd_ff,nFd_Df*1e3,nFd_rmse*100));


         double nd_fF = fitFluid(b,y,DF,noiseFloor);
         double nd_ff = 1-nd_fF;
         double nd_Df = Double.NaN;
         if (nd_fF < 1) nd_Df = fitFlow(b,arrNorm(arrDiff(y,se(b,nd_fF,DF,1))));
         double nd_rmse = rmse(by[1],arrAdd(se(by[0],nd_fF,DF,aF),se(by[0],nd_ff,nd_Df,1)));

         plot.setColor(Color.magenta);
         plot.setLineWidth(1);
         plot.addPoints(b,arrln(arrAdd(se(b,nd_fF,DF,aF),se(b,nd_ff,nd_Df,1))),Plot.LINE);
         plot.addLabel(0.01,0.71+2*lh,String.format("Fluid         %5.2f %7.2f %5.2f                   %5.2f",nd_ff,nd_Df*1e3,nd_fF,nd_rmse*100));

         plot.setLineWidth(3);
         plot.setColor(Color.red);
         plot.addPoints(new double[]{1500},arrln(se(new double[]{1500},1,3e-3,1)),Plot.CROSS);
         plot.setLineWidth(1);
         

         double slse_ff = 0;
         double slse_Df = Double.NaN;
         double slse_fF = 0;
         double slse_fd = 0;
         double slse_Dd = Double.NaN;
         double slse_ad = Double.NaN;
         double slse_af = Double.NaN;
         double slse_rmse = Double.NaN;
         //double[] fdDdad = fitSubLinearSE(b,y,new double[]{1200,b[b.length-1]}, noiseFloor);
         double[] fdDdad = fitSEFraction(b,y,new double[]{2000,4000}, new double[]{1500,3000}, noiseFloor);
         if (fdDdad != null && fdDdad[0]<1.2 && fdDdad[1]<3e-3 && fdDdad[2]>0.5 && fdDdad[2]<1.1) {
            slse_fd = fdDdad[0];
            slse_Dd = fdDdad[1];
            slse_ad = fdDdad[2];
            double[] slse_y2 = arrNorm(arrDiff(y,se(b,slse_fd,slse_Dd,1)));
            slse_fF = (1-slse_fd)*fitFluid(b,slse_y2,DF,noiseFloor);

            double[] slse_y3 = arrNorm(arrDiff(y,arrAdd(se(b,slse_fd,slse_Dd,1),se(b,slse_fF,DF,1))));
   
            slse_ff = 1 - slse_fF - slse_fd;
            slse_Df = Double.NaN;
            slse_af = 1;
            if (slse_ff < 0.01) slse_ff = 0;
            else slse_Df = fitFlow(b,slse_y3);

            if ((slse_ff > 0) && Double.isNaN(slse_Df))
               slse_Df = 1;
/*
            plot.setColor(Color.green);
            plot.setLineWidth(2);
            plot.addPoints(b,arrln(arrAdd(se(b,slse_fF,DF,aF),se(b,slse_ff,slse_Df,1))),Plot.LINE);
*/
            }

         double[] slse_y2 = arrNorm(arrDiff(y,se(b,slse_fd,slse_Dd,1)));
         slse_fF = (1-slse_fd)*fitFluid(b,slse_y2,DF,noiseFloor);

         double[] slse_y3 = arrNorm(arrDiff(y,arrAdd(se(b,slse_fd,slse_Dd,1),se(b,slse_fF,DF,aF))));
         slse_ff = 1 - slse_fF - slse_fd;
         slse_Df = Double.NaN;
         slse_af = 1;
         if (slse_ff < 0.01) slse_ff = 0;
         else slse_Df = fitFlow(b,slse_y3);

         if ((slse_ff > 0) && Double.isNaN(slse_Df))
            slse_Df = 1;

         slse_rmse = rmse(by[1],arrAdd(se(by[0],slse_fd,slse_Dd,slse_ad),se(by[0],slse_fF,DF,aF),se(by[0],slse_ff,slse_Df,slse_af)));
         plot.setLineWidth(2);
         plot.setColor(Color.red);
         plot.addPoints(b,arrln(arrAdd(se(b,slse_fd,slse_Dd,slse_ad),se(b,slse_fF,DF,aF),se(b,slse_ff,slse_Df,slse_af))),Plot.LINE);
         plot.addLabel(0.01,0.71+3*lh,String.format("Diffusion(se) %5.2f %7.2f %5.2f %5.2f %5.2f %5.2f %5.2f",slse_ff,slse_Df*1e3,slse_fF,slse_fd,slse_Dd*1e3,slse_ad,slse_rmse*100));


         double le_fd = 0;
         double le_Dd = Double.NaN;
         double le_ad = 1.0;
         double[][] diffby = arrDeNaN(arrSubSet(b,y,new double[]{1200, 5000}, noiseFloor));

         plot.setColor(Color.blue);

         if ((diffby[0].length > 1) && (arrMax(diffby[0]) > 1500)) {
            plot.setColor(Color.blue);
            plot.setLineWidth(3);
            plot.addPoints(diffby[0],arrln(diffby[1]),Plot.X);

            double[] mb = theilsen(diffby[0],arrln(diffby[1]));
            le_fd = Math.exp(mb[1]);
            le_Dd = -mb[0];
                 if ((le_Dd > DF) || (le_Dd < 0) || Double.isNaN(le_fd)) {
               le_fd = 0;
               le_Dd = Double.NaN;
               //Dd = DF;
               //fd = Math.exp(theilsenB(diffby[0],arrln(diffby[1]),-Dd));
               }
            else if (le_fd > 1) {
               double[][] diff2by = arrDeNaN(arrSubSet(b,y,new double[]{300, 2500}, noiseFloor));
               le_fd = 1;
               le_Dd = -centerM(diff2by[0],arrln(diff2by[1]));
               plot.setColor(Color.cyan);
/*
               plot.setLineWidth(1);
               plot.addPoints(b,arrln(se(b,le_fd,le_Dd,le_ad)),Plot.LINE);
*/
               }
            else {
               plot.setColor(Color.blue);
/*
               plot.setLineWidth(1);
               plot.addPoints(b,arrln(se(b,le_fd,le_Dd,le_ad)),Plot.LINE);
*/
               }
            }
      else {
         double[][] diff2by = arrDeNaN(arrSubSet(b,y,new double[]{300, 2500}, noiseFloor));
         double[] mb = theilsen(diff2by[0],arrln(diff2by[1]));
         le_fd = Math.exp(mb[1]);
         le_Dd = -mb[0];
         if ((le_Dd > 3e-3) || (le_Dd < 0) || Double.isNaN(le_fd)) {
            le_fd = 0;
            le_Dd = Double.NaN;
            }
         }

         double[] le_y2 = arrNorm(arrDiff(y,se(b,le_fd,le_Dd,1)));
         double le_fF = (1-le_fd)*fitFluid(b,le_y2,DF,noiseFloor);
/*
         if (le_fF > 0) {
            double[][] fluidby = arrDeNaN(arrSubSet(b,le_y2,new double[]{300, 1500},noiseFloor));
            plot.setColor(Color.red);
            plot.setLineWidth(2);
            plot.addPoints(fluidby[0],arrln(fluidby[1]),Plot.X);
            }
*/
         double[] y3 = arrNorm(arrDiff(y,arrAdd(se(b,le_fd,le_Dd,1),se(b,le_fF,DF,aF))));
         double[][] flowby = arrSubSet(b,y3,new double[]{0,150});
/*
         plot.setColor(Color.green);
         plot.addPoints(flowby[0],arrln(flowby[1]),Plot.X);
*/
         double le_ff = 1 - le_fF - le_fd;
         double le_Df = Double.NaN;
         double le_af = 1;
         if (le_ff < 0.01) le_ff = 0;
         else le_Df = fitFlow(b,y3);

         if ((le_ff > 0) && Double.isNaN(le_Df))
            le_Df = 1;

         double le_rmse = rmse(by[1],arrAdd(se(by[0],le_fd,le_Dd,le_ad),se(by[0],le_fF,DF,aF),se(by[0],le_ff,le_Df,le_af)));
         plot.addPoints(b,arrln(arrAdd(se(b,le_fd,le_Dd,le_ad),se(b,le_fF,DF,aF),se(b,le_ff,le_Df,le_af))),Plot.LINE);
         plot.addLabel(0.01,0.71+4*lh,String.format("Diffusion(me) %5.2f %7.2f %5.2f %5.2f %5.2f %5.2f %5.2f",le_ff,le_Df*1e3,le_fF,le_fd,le_Dd*1e3,le_ad,le_rmse*100));


         double ff = le_ff;
         double Df = le_Df;
         double af = le_af;

         double fd = 0;
         double Dd = Double.NaN;
         double ad = 1.0;
         double fF = 0;

         fF = le_fF;
         fd = le_fd;
         Dd = le_Dd;
         ad = le_ad;

/*
         plot.setLineWidth(2);
         plot.setColor(Color.green);
         plot.addPoints(b,arrln(arrAdd(se(b,fd,Dd,ad),se(b,fF,DF,aF),se(b,ff,Df,af))),Plot.LINE);
         plot.setColor(Color.red);
         plot.addPoints(b,arrln(arrAdd(se(b,fd,Dd,ad),se(b,fF,DF,aF))),Plot.LINE);
         plot.setColor(Color.blue);
         plot.addPoints(b,arrln(se(b,fd,Dd,ad)),Plot.LINE);


         plot.setColor(Color.black);
         plot.setJustification(Plot.RIGHT);
         plot.addLabel(0.6, 0.1, " \nf\nD");

         plot.setColor(Color.blue);
         plot.addLabel(0.675, 0.1, "Diff\n"+d2s(fd)+"\n"+d2s(Dd*1e3));
         plot.setColor(Color.red);
         plot.addLabel(0.770, 0.1, "Fluid\n"+d2s(fF)+"\n"+d2s(DF*1e3));
         plot.setColor(Color.green);
         plot.addLabel(0.870, 0.1, "Flow\n"+d2s(ff)+"\n"+d2s(Df*1e3));

         plot.setColor(Color.black);
         plot.setJustification(Plot.LEFT);
         plot.addLabel(0.875, 0.1, " \n\u00b5m\u00b2/ms\n ");
*/
         plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));

         switch(arrMinAt(new double[]{nd_rmse,nFd_rmse,slse_rmse,le_rmse})) {
            case 0: ff=nd_ff;   Df=nd_Df;   fF=nd_fF;   fd=0;       Dd=Double.NaN; ad=Double.NaN; break;
            case 1: ff=nFd_ff;  Df=nFd_Df;  fF=0;       fd=0;       Dd=Double.NaN; ad=Double.NaN; break;
            case 2: ff=slse_ff; Df=slse_Df; fF=slse_fF; fd=slse_fd; Dd=slse_Dd;    ad=slse_ad;    break;
            case 3: ff=le_ff;   Df=le_Df;   fF=le_fF;   fd=le_fd;   Dd=le_Dd;      ad=1;          break;
            }
/*
         if (nd_rmse < le_rmse) {
            ff = nd_ff;
            Df = nd_Df;
            fF = nd_fF;
            fd = 0;
            Dd = Double.NaN;
            // update noiseFloor
            plot.setColor(Color.red);
            plot.addLabel(0.5,0.5,"Override");
            }
*/

         //plot.setLimitsToFit(true);
         plot.setLimits(0,2500,-5,0);
         pwv.get(n).drawPlot(plot);
         return new double[]{ff,Df,fF,DF,fd,Dd,noiseFloor};
         }
      catch(Exception e) {
         StringWriter sw = new StringWriter();
         e.printStackTrace(new PrintWriter(sw));
         IJ.error("An error occurred.\n\n"+sw.toString());
         }


      plot.setLimitsToFit(true);
      pwv.get(n).drawPlot(plot);
      return null;
      }
      
   public static double[] fitNonLinSE(double[] b, double[] y) {
      CurveFitter cf = new CurveFitter(b,y);
      cf.setMaxIterations(2000);
      cf.doCustomFit(new UserFunction() {
         @Override
         public double userFunction(double[] p, double x) { return se(x,p[0],p[1]); }
         }, 2, "", new double[]{1e-3,1}, new double[]{0.1*1e-3,0.1*1}, false);
      double[] p = cf.getParams();
      //p[2] = cf.getRSquared();
      //double[] yy = arrSum(se(b,ff,Df,af),se(b,fd,Dd,ad));
      double[] yy = se(b,1,p[0],p[1]);
      p[2] = Math.sqrt(arrMeanSquareDiff(arrNorm(y),yy));
      return p;
      }

   public static double[] fitIVIMSE(double[] b, double[] y, double noiseFloor) {
      double[] pIVIM = fitIVIM(b,y);
      double ff = pIVIM[0];
      double Df = pIVIM[1];

      double[] pSE = fitLinSE(b, arrNorm(arrDiff(y,se(b,ff,Df,1))), noiseFloor);
      double fd = 1-ff;
      double Dd = pSE[1];
      double ad = pSE[2];

      double[] yy = arrSum(se(b,ff,Df,1),se(b,fd,Dd,ad));
      double rms = Math.sqrt(arrMeanSquareDiff(arrNorm(y),yy));

      return new double[]{ff,Df,fd,Dd,ad,rms};
      }

   public static double fitOnlyFluid(double[] b, double[] y) {
      return rmse(y,se(b,1,3e-3,1));
      }

   public static double[] fitLinME(double[] b, double[] y, double[] range, double noiseFloor) {
      double[][] by = arrDeNaN(arrSubSet(b,y,range, noiseFloor));
      double[] mb = theilsen(by[0],arrln(by[1]));
      double f = Math.exp(mb[1]);
      double D = -mb[0];
      double[] yy = se(b,f,D,1);
      double rmse = rmse(y,yy);
      return new double[]{f,D,rmse};
      }

   public static double[] fitLinSE(double[] b, double[] y, double[] range, double noiseFloor) {
      double[][] by = arrDeNaN(arrSubSet(b,y,range, noiseFloor));
      double[][] ls = arrDeNaN(linse(by));

      double[] mb = theilsen(ls);
      double fd = 1;
      double ad = mb[0];
      double Dd = Math.exp(mb[1]/ad);
      
      double[] yy = se(b,fd,Dd,ad);
      double rms = rmse(y,yy);

      return new double[]{fd,Dd,ad,rms};
      }
   public static double[] fitLinSE(double[] b, double[] y, double noiseFloor) {

      double[][] diffby = arrDeNaN(arrSubSet(b,y,new double[]{500, 4000}, noiseFloor));
      double[][] diffls = arrDeNaN(linse(diffby));

      double[] mb = theilsen(diffls);
      double fd = 1;
      double ad = mb[0];
      double Dd = Math.exp(mb[1]/ad);

      if (Dd > 3e-3) {  // Need to loop until acceptable or out of data points
         double[] ytmp = arrDeNaN(y);
         noiseFloor = ytmp[ytmp.length-1];

         double[][] diff2by = arrDeNaN(arrSubSet(b,y,new double[]{500, 4000}, noiseFloor));
         double[][] diff2ls = arrDeNaN(linse(diff2by));
         double[] mb2 = theilsen(diff2ls);
         ad = mb2[0];
         Dd = Math.exp(mb2[1]/ad);
         }

      //double[] yy = arrSum(se(b,ff,Df,af),se(b,fd,Dd,ad));
      double[] yy = se(b,fd,Dd,ad);
      //double rms = Math.sqrt(arrMeanSquareDiff(arrNorm(y),yy));
      double rms = rmse(y,yy);

      return new double[]{fd,Dd,ad,rms};
      }


   private double[] plotLinSE(int n, int sx, int sy, double[] b, double[] y, double noiseFloor) {
      PlotWindow.noGridLines = false; // draw grid lines
      Plot plot = new Plot("Linerized Stretched Exponential","ln b-value (ms/um^2)","ln(y_i/y_0)");
      
      plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
      plot.setXLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
      plot.setYLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
      plot.setLineWidth(2);
      plot.setWindowSize(3*dx, 1*dy);
      
      if ( (pwv.size() < (n+1)) || (pwv.get(n) == null)  || pwv.get(n).isClosed()) {
         Plot p = new Plot("Linerized Stretched Exponential","","");
         p.setWindowSize(3*dx,1*dy);
         pwv.add(n,p.show());
         //pwv.get(n).setSize(3*dx,1*dy);
         //pwDim = pwv.get(n).getSize();
         pwv.get(n).setLocation(sx,sy);
         pwv.get(n).windowActivated(null);
         }

      try {
         double[][] ls = arrDeNaN(linse(b,y));
//if (ls[0].length == 0) IJ.log("ls[0] is empty");
//if (ls[1].length == 0) IJ.log("ls[1] is empty");
//if ((ls[0].length == 0) || (ls[1].length == 0)) arrLog(b,y);
         plot.setLineWidth(2);
         plot.addPoints(ls[0],ls[1],Plot.X);

         double[][] diffby = arrDeNaN(arrSubSet(b,y,new double[]{500, 4000}, noiseFloor));
         double[][] diffls = arrDeNaN(linse(diffby));
         plot.setLineWidth(3);
         plot.addPoints(diffls[0],diffls[1],Plot.X);

         plot.setFont(new Font("Comic Sans MS", Font.BOLD, 14));
         plot.setColor(Color.black);
         plot.setJustification(Plot.RIGHT);
         plot.addLabel(0.7, 0.8, "Dd\nalpha");

         double[] mb = theilsen(diffls);
         plot.setLineWidth(1);
         plot.addPoints(ls[0],makeLine(ls[0],mb),Plot.LINE);
         plot.setLineWidth(2);
         plot.addPoints(diffls[0],makeLine(diffls[0],mb),Plot.LINE);
         double fd = 1;
         double ad = mb[0];
         double Dd = Math.exp(mb[1]/ad);

         plot.setColor(Color.blue);
         plot.addLabel(0.775, 0.8, String.format("%5.2f",Dd*1e3)
                             +"\n"+String.format("%5.2f",ad));


         plot.setLineWidth(2);

         if (Dd > 3e-3) {
            double[] ytmp = arrDeNaN(y);
            noiseFloor = Math.log(ytmp[ytmp.length-1]);

            plot.setColor(Color.red);
            double[][] diff2by = arrDeNaN(arrSubSet(b,y,new double[]{500, 4000}, Math.exp(noiseFloor)));
//arrLog(diff2by);
            double[][] diff2ls = arrDeNaN(linse(diff2by));
            plot.setLineWidth(3);
            plot.addPoints(diff2ls[0],diff2ls[1],Plot.X);

            double[] mb2 = theilsen(diff2ls);
            plot.setLineWidth(1);
            plot.addPoints(ls[0],makeLine(ls[0],mb2),Plot.LINE);
            plot.setLineWidth(2);
            plot.addPoints(diff2ls[0],makeLine(diff2ls[0],mb2),Plot.LINE);

            ad = mb2[0];
            Dd = Math.exp(mb2[1]/ad);
            plot.addLabel(0.875, 0.8, String.format("%5.2f",Dd*1e3)
                                +"\n"+String.format("%5.2f",ad)
                                +"\n"+String.format("%5.2f",noiseFloor)+"  "+String.format("%5.2f",Math.exp(noiseFloor)));
            }

        

         plot.setColor(Color.black);
         plot.setJustification(Plot.LEFT);
         plot.addLabel(0.875, 0.8, " \u00b5m\u00b2/ms\n ");
         plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));

         plot.setLimitsToFit(true);
         //plot.setLimits(0,2500,-5,0);
         pwv.get(n).drawPlot(plot);
         return new double[]{fd,Dd,ad,noiseFloor};
         }
      catch(Exception e) {
         StringWriter sw = new StringWriter();
         e.printStackTrace(new PrintWriter(sw));
         IJ.error("An error occurred.\n\n"+sw.toString());
         }

      pwv.get(n).drawPlot(plot);
      plot.setLimitsToFit(true);
      return null;
      }
      

   public static double[] arrMul(double[] arr, double a) {
      double[] ans = new double[arr.length];
      for(int i=0; i<arr.length; i++)
         ans[i] = a*arr[i];
      return ans;
      }

   public static void arrLog(String name, double[] arr) {
      IJ.log(name);
      for(double a : arr)
         IJ.log(""+a);
      }
   public static void arrLog(double[] arr) {
      for(double a : arr)
         IJ.log(""+a);
      }
   public static void arrLog(double[] x, double[] y) {
      for(int i=0; i<x.length; i++)
         IJ.log(""+x[i]+"   "+y[i]);
      }
   public static void arrLog(double[][] arr) {
      IJ.log("arr.length="+arr.length);
      IJ.log("arr[0].length="+arr[0].length);
      IJ.log("arr[1].length="+arr[1].length);
      for(int i=0; i<arr[0].length; i++)
         IJ.log(""+arr[0][i]+"    "+arr[1][i]);
      }
////////

   public static double[] fitSEFraction(double[] b, double[] y, double[] linfitRange, double[] testRange, double noiseFloor) {
         double F = 1.0;
         double D = Double.NaN;
         double a = Double.NaN;
      try {
         double rmseMin = Double.POSITIVE_INFINITY;
         for(double f=1.0,df=0.05; f>0.0; f-=df) {
            double[] yg = arrDiv(y,f);
            double[][] lslf = linse(arrSubSet(b,yg,linfitRange, noiseFloor));
            double[] mb = theilsen(lslf);
            double rmse = rmse(arrSubSet(b,yg,testRange),arrSubSet(b,se(b,f,Math.exp(mb[1]/mb[0]),mb[0]),testRange));
IJ.log("f="+f+"  D="+Math.exp(mb[1]/mb[0])+"  a="+mb[0]+"  rmse="+rmse);
            if (Double.isNaN(rmse))
               return new double[]{F,D,a};
            if (rmse < rmseMin) {
               rmseMin = rmse;
               F = f;
               a = mb[0];
               D = Math.exp(mb[1]/a);
IJ.log("min found");
               }
            }
IJ.log("F="+F+"  D="+D+"  a="+a);
         return new double[]{F,D,a};
         }
      catch(ArrayIndexOutOfBoundsException e) {
         return new double[]{F,D,a};
         }   
      catch(Exception e) {
         StringWriter sw = new StringWriter();
         e.printStackTrace(new PrintWriter(sw));
         IJ.log("fitSEFraction: AN error occurred.\n\n"+sw.toString());
         return null;
         }   
      }

   public static double fitMonoFraction2(double[] b, double[] y, double D, double alpha, double[] linfitRange, double[] testRange, double noiseFloor) {
      try {
         double[][] by = arrDeNoise(b,y,Math.log(noiseFloor));
         if ((by[0].length < 2) || (by[0][by[0].length-1] < linfitRange[0]))
            return 0;
         double F = 0F;
         double cmin = Double.POSITIVE_INFINITY;
         for(double f=0, df=0.005; f<1; f+=df) {
            double c = arrLinearTest(by[0], by[1], f, D, alpha, linfitRange, testRange, noiseFloor);
            if (Double.isNaN(c))
               return F;
            if (c < cmin) {
               cmin = c;
               F = f;
               }
            }
         double dF = 0.002F;
         int[] DRange = arrElementNear(by[0],linfitRange);
         int[] RRange = arrElementNear(by[0],testRange);
         double f = nextGuess(by[0], by[1], dF, F-dF, D, alpha, linfitRange, testRange, noiseFloor);
              if (Double.isNaN(f)) f = F;
         else if (f<0)             f = 0;
         else if (f>1)             f = 0;
         if (Double.isNaN(f))
            f = 0;

         return f;
         }
      catch(ArrayIndexOutOfBoundsException e) {
         return 0;
         }   
      catch(Exception e) {
         StringWriter sw = new StringWriter();
         e.printStackTrace(new PrintWriter(sw));
         IJ.log("fitMonoFraction: AN error occurred.\n\n"+sw.toString());
         return 0;
         }   
      }

   public static double arrLinearTest(double[] b, double[] signal, double f, double D, double alpha, double[] rl, double[] rt, double noiseFloor) {
      double[][] by  = arrSubSet(b,arrDiff(signal,se(b,f,D,alpha)),rt,noiseFloor);
      double[] sig = arrNorm(arrDiff(signal,se(b,f,D,alpha)));
      double[] mb  = theilsen(linse(arrSubSet(b,sig,rl)));
      return (by[0].length == 0) ? Double.NaN : arrSumSquareDiff(arrSubSet(b,sig,rt),arrSubSet(b,se(b,1,Math.exp(mb[1]/mb[0]),mb[0]),rt));
      }

   public static double nextGuess(double[] b, double[] signal, double df, double f, double D, double alpha, double[] rl, double[] rt, double noiseFloor) {
      double ym1 = arrLinearTest(b, signal, f-df, D, alpha, rl, rt, noiseFloor);
      double y   = arrLinearTest(b, signal, f   , D, alpha, rl, rt, noiseFloor);
      double yp1 = arrLinearTest(b, signal, f+df, D, alpha, rl, rt, noiseFloor);
   
      double ymp = (y-ym1)/df;
      double ypp = (yp1-y)/df;
   
      double M = (ypp-ymp)/df;
      double B = y - M*(f-df);
      return -B/M;
      }

////////

   public static double[] fitIVIM(double[] b, double[] y) {

      // Assume signal between b=500 to b=1500 is linear
      int bi500  = arrElementNear(b, 500);
      int bi1500 = arrElementNear(b,1500);
      double[][] arrlin = arrln(arrDeNoise(arrSubSet(b,arrNorm(y),new int[]{bi500-1,bi1500}),-3.5));
      double[] mb = theilsen(arrlin);
      double fd = Math.exp(mb[1]);
      double Dd = -mb[0];
      double ad = 1;
      if (Dd > 3.0E-3F) {
         Dd = 3E-3F;
         fd = Math.exp(theilsenB(arrlin[0],arrlin[1],-Dd));
         }

      // Assume other signal below b=500 is pseudo diffusion (flow)
      double ff = 1F - fd;
      double Df = 0F;
      double af = 1;
      try {
         double[] jf = arrSubSet(arrln(arrNorm(arrDiff(arrNorm(y),se(b,fd,Dd,1F)))),new int[]{0,bi500-1});
         double[][] ybf = arrDeNaN(arrSubSet(b,new int[]{0,bi500-1}),jf);

          if (ybf[0].length > 1) {
             double[] dfydb = new double[ybf[0].length-1];
             for(int i=0; i<dfydb.length; i++)
                dfydb[i] = -ybf[1][i+1]/ybf[0][i+1];
             Df = center(dfydb);
             if (Df < 3E-3F) Df = arrMax(dfydb);
             if (Df < 3E-3F) Df = 3E-3F;
             }
          else {
             ff = 0F;
             Df = 0F;
             }


         if ((ff<0F) || Double.isNaN(Df) || (Df<3E-3F)) {
            ff = 0F;
            Df = Double.NaN;
            }

         }
      catch(Exception e) {
         ff = 0F;
         Df = 0F;
         }

      double[] yy = arrSum(se(b,ff,Df,af),se(b,fd,Dd,ad));
      double rms = Math.sqrt(arrMeanSquareDiff(arrNorm(y),yy));

      return new double[]{ff, Df, fd, Dd, rms};
    }


/////////////////////////////

    // these listeners are activated if the selection is changed in the corresponding ImagePlus
    public synchronized void mousePressed(MouseEvent e) {  }   
    public synchronized void mouseDragged(MouseEvent e) {  }
    public synchronized void mouseClicked(MouseEvent e) { doPlotting(((ImageCanvas)(e.getSource())).getImage()); }
    public synchronized void keyPressed(KeyEvent e)     { doPlotting(((ImageCanvas)(e.getSource())).getImage()); }
    // unused listeners concering actions in the corresponding ImagePlus
    public void mouseReleased(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseMoved(MouseEvent e) {}
    public void keyTyped(KeyEvent e) {}
    public void keyReleased(KeyEvent e) {}

    public void windowActivated(WindowEvent e)   {}
    public void windowClosed(WindowEvent e)      { disengage((ImageWindow)e.getWindow());}
    public void windowClosing(WindowEvent e)     { disengage((ImageWindow)e.getWindow());}
    public void windowDeactivated(WindowEvent e) {}
    public void windowDeiconified(WindowEvent e) {}
    public void windowIconified(WindowEvent e)   {}
    public void windowOpened(WindowEvent e)      {}

//////////////////////////////////////////////////////////////////////////////
   public static float[] d2f(double[] x) {
      float[] res = new float[x.length];
      for(int i=0; i<res.length; i++)
         res[i] = (float)x[i];
      return res;
      }

   public static double se(double b, double D, double alpha) {
      return Math.exp(-Math.pow(b*D,alpha));
      }
   public static double[] se(double[] b, double f, double D, double alpha) {
      double[] ans = new double[b.length];
      if (f==0)
         return ans;
      for(int i=0; i<b.length; i++)
         ans[i] = f*(double)Math.exp(-Math.pow(b[i]*D,alpha));
      return ans;
      }
   public static double[][] linse(double[][] arr) {
      return linse(arr[0],arr[1]);
      }
   public static double[][] linse(double[] b, double[] s) {
      if (b.length < 1) return new double[2][0];
      // ignore b=0
      int igb0 = (b[0] < 1E-2) ? 1 : 0;
      double[][] ans = new double[2][b.length-igb0];
      for(int i=0; i<(b.length-igb0); i++) {
         ans[0][i] = Math.log(b[i+igb0]);
         ans[1][i] = Math.log(-Math.log(s[i+igb0]));
         }
      return ans;
      }

   public static int arrElementNear(double[] arr, double v) {
      int ei = 0;
      double ed = Double.POSITIVE_INFINITY;
      for(int i=0; i<arr.length; i++)
         if (Math.abs(arr[i]-v) < ed) {
            ed = Math.abs(arr[i]-v);
            ei = i;
            }
      return ei;
      }
   public static int[] arrElementNear(double[] arr, double[] vs) {
      int[] res = new int[vs.length];
      for(int i=0; i<vs.length; i++)
         res[i] = arrElementNear(arr,vs[i]);
      return res;
      }

   public static double[] arrDeNaN(double[] y) {
      int n = 0;
      for(int i=0; i<y.length; i++)
         if (!(Double.isNaN(y[i])||Double.isInfinite(y[i])))
            n++;
      double[] res = new double[n];
      for(int i=0,j=0; i<y.length; i++)
         if (!(Double.isNaN(y[i])||Double.isInfinite(y[i]))) {
            res[j] = y[i];
            j++;
            }
      return res;
      }
   public static double[][] arrDeNaN(double[] x, double[] y) {
      int n = 0;
      for(int i=0; i<y.length; i++)
         if (!(Double.isNaN(y[i])||Double.isInfinite(y[i])))
            n++;
      double[][] res = new double[2][n];
      if (n==0) return res;
      for(int i=0,j=0; i<y.length; i++)
         if (!(Double.isNaN(y[i])||Double.isInfinite(y[i]))) {
            res[0][j] = x[i];
            res[1][j] = y[i];
            j++;
            }
      return res;
      }
   public static double[][] arrDeNaN(double[][] arr) {
      int n = 0;
      for(int i=0; i<arr[0].length; i++)
      if (!(Double.isNaN(arr[1][i])||Double.isInfinite(arr[1][i])))
         n++;
   double[][] res = new double[2][n];
   for(int i=0,j=0; i<arr[0].length; i++)
      if (!(Double.isNaN(arr[1][i])||Double.isInfinite(arr[1][i]))) {
         res[0][j] = arr[0][i];
         res[1][j] = arr[1][i];
         j++;
         }
   return res;
   }

   public static double[][] arrDeNoise(double[] x, double[] y, double noise) {
      int n = 0;
      for(n=0; n<y.length; n++) 
         if (Math.log(y[n]) < noise)
            break;
      double[][] res = new double[2][n];
      for(int i=0; i<n; i++) {
         res[0][i] = x[i];
         res[1][i] = y[i];
         }
      return res;
      }
   public static double[][] arrDeNoise(double[][] arr, double noise) {
      return arrDeNoise(arr[0], arr[1], noise);
      }

   public static void arrsLog(double[] x, double[] y) {
      for(int i=0; i<x.length; i++)
         IJ.log(""+i+"  "+x[i]+"  "+y[i]);
      }

   public static double[] makeLine(double[] x, double m, double b) {
      double[] y = new double[x.length];
      for(int i=0; i<x.length; i++)
         y[i] = m*x[i] + b;
      return y;
      }
   public static double[] makeLine(double[] x, double[] mb) {
      return makeLine(x,mb[0],mb[1]);
      }

   public static double[][] arrln(double[][] a) {
      double[][] ans = new double[2][a[0].length];
      for(int i=0; i<a[0].length; i++) {
         ans[0][i] = a[0][i];
         ans[1][i] = Math.log(a[1][i]);
         }
      return ans;   
      }
   public static double[] arrln(double[] a) {
      double[] ans = new double[a.length];
      for(int i=0; i<a.length; i++) {
         ans[i] = Math.log(a[i]);
         }
      return ans;
      }
   public static double[] arrAdd(double[] ... arrs) {
      double[] ans = new double[arrs[0].length];
      for(double[] arr : arrs)
         for(int i=0; i<arrs[0].length; i++)
            ans[i] += arr[i];
      return ans;
      }
   public static double[] arrDiv(double[] arr, double d) {
      double[] ans = new double[arr.length];
      for(int i=0; i<arr.length; i++)
         ans[i] = arr[i]/d;
      return ans;
      }
   public static double[] arrSub(double[] arr, double s) {
      double[] ans = new double[arr.length];
      for(int i=0; i<arr.length; i++)
         ans[i] = arr[i] - s;
      return ans;
      }
   public static double[] arrDiff(double[] a, double[] b) {
      double[] ans = new double[a.length];
      for(int i=0; i<a.length; i++) {
         ans[i] = a[i] - b[i];
         }
      return ans;
      }
   public static double arrMax(double[] a) {
      if (a.length == 0)
         return Double.NaN;
      double max = a[0];
      for(int i=0; i<a.length; i++)
         if (a[i] > max) max = a[i];
      return max;
      }
   public static int arrMaxAt(double[] a) {
      int maxat = 0;
      for(int i=0; i<a.length; i++)
         if (a[i] > a[maxat]) maxat = i;
      return maxat;
      }
   public static double arrMin(double[] a) {
      if (a.length == 0)
         return Double.NaN;
      double min = a[0];
      for(int i=0; i<a.length; i++)
         if (a[i] < min) min = a[i];
      return min;
      }
   public static int arrMinAt(double[] a) {
      int minat = 0;
      for(int i=0; i<a.length; i++)
         if (a[i] < a[minat]) minat = i;
      return minat;
      }
   public static double[] arrNorm(double[] a) {
      double max = arrMax(a);
      double[] ans = new double[a.length];
      for(int i=0; i<a.length; i++)
         ans[i] = a[i]/max;
      return ans;
      }
   public static double[] arrNorm(double[] a, double min) {
      double max = arrMax(a);
      double[] ans = new double[a.length];
      for(int i=0; i<a.length; i++)
         ans[i] = (a[i]>min) ? a[i]/max : Double.NaN;
      return ans;
      }
   public static double[] arrNorm(double[] a, float b0v, String ugh) {
      double[] ans = new double[a.length];
      for(int i=0; i<a.length; i++)
         ans[i] = a[i]/b0v;
      ans[0] = 1;
      return ans;
      }
   public static double[] arrSquareDiff(double[] a, double[] b) {
      double[] ans = new double[a.length];
      for(int i=0; i<a.length; i++)
         ans[i] = (a[i]-b[i])*(a[i]-b[i]);
      return ans;
      }
   public static double arrMean(double[] a) {
      double ans = 0;
      for(int i=0; i<a.length; i++)
         ans += a[i];
      return ans/a.length;   
      }
   public static double arrMean(double[][] a) {
      return arrMean(a[1]);
      }
   public static double arrSumSquareDiff(double[][] a, double[][] b) {
      return arrSumSquareDiff(a[1],b[1]);
      }
   public static double arrSumSquareDiff(double[] a, double[] b) {
      if (a.length < 1) 
         return Double.NaN;
      double ans = 0;
      for(int i=0; i<a.length; i++)
         ans += (a[i]-b[i])*(a[i]-b[i]);
      return ans;
      }
   public static double rmse(double[] a, double[] b) {
      return Math.sqrt(arrMeanSquareDiff(a,b));
      }
   public static double rmse(double[][] a, double[][] b) {
      return rmse(a[1],b[1]);
      }
   public static double arrMeanSquareDiff(double[][] a, double[][] b) {
      return arrMeanSquareDiff(a[1],b[1]);
      }
   public static double arrMeanSquareDiff(double[] a, double[] b) {
      return arrSumSquareDiff(a,b)/a.length;
      }
   public static double arrMeanSquareDiff(double[] a, double[] b, int n) {
      double[] sda = arrSquareDiff(a,b);
      for(int i=0; i<n; i++) 
         sda[arrMaxAt(sda)] = Double.NaN;
      double sum = 0F;
      for(int i=0; i<sda.length; i++)
         if (!Double.isNaN(sda[i])) sum += sda[i];
      return sum/(sda.length-n);
      }
   public static double[][] arrSubSet(double[][] a, int from) {
      if (a[0].length-from < 1) return new double[2][0];
      double[][] ans = new double[2][a[0].length-from];
      for(int i=0; i<ans[0].length; i++) {
         ans[0][i] = a[0][i+from];
         ans[1][i] = a[1][i+from];
         }
      return ans;
      }
   public static double[] arrSubSet(double[] a, int from) {
      double[] ans = new double[a.length-from];
      for(int i=0; i<ans.length; i++)
         ans[i] = a[i+from];
      return ans;
      }
   public static double[][] arrSubSet(double[] x, double[] y, int[] r) {
      if (r[1]<r[0]) return new double[2][0];
      double[][] ans = new double[2][r[1]-r[0]+1];
      for(int i=0; i<ans[0].length; i++) {
         ans[0][i] = x[i+r[0]];
         ans[1][i] = y[i+r[0]];
         }
      return ans;
      }
   public static double[][] arrSubSet(double[][] a, int[] r) {
      return arrSubSet(a[0], a[1], r);
      }
   public static double[] arrSubSet(double[] a, int[] range) {
      double[] ans = new double[range[1]-range[0]+1];
      for(int i=0; i<ans.length; i++)
         ans[i] = a[i+range[0]];
      return ans;
      }

   public static double[][] arrSubSet(double[][] xy, double[] r) {
      return arrSubSet(xy[0],xy[1],r);
      }
   public static double[][] arrSubSet(double[] x, double[] y, double[] r) {
      if (x.length < 1) return new double[2][0];
      int s = arrElementNear(x,r[0]);
      int e = arrElementNear(x,r[1]);
      return arrSubSet(x,y,new int[]{s,e});
      }

   public static int arrElementGE(double[] arr, double v) {
      for(int i=0; i<arr.length; i++)
         if (arr[i] < v)
            return i-1;
      return arr.length;
      }

   public static double[][] arrSubSet(double[] x, double[] y, double[] r, double noise) {
      int s = arrElementNear(x,r[0]);
      int e = Math.min(arrElementNear(x,r[1]),arrElementGE(y,noise));
      return arrSubSet(x,y,new int[]{s,e});
      }

   public static int countAbove(double[] a, double[] b) {
      int count = 0;
      for(int i=0; i<a.length; i++)
         if (a[i]>b[i]) count++;
      return count;
      }

//////////////////////////////////////////////////////////////////////////////
   public static double median(double[] A) {  // A is modified
      if (A.length <  1) return Double.NaN;
      if (A.length == 1) return A[0];
      Arrays.sort(A);
      int i = A.length/2;
      return (A.length%2 == 1) ? A[i] : (A[i-1]+A[i])/2.0;
      }
   public static double center(double[] A) {  // A is modified
      double m  = median(A);
      if (A.length < 4) return m;
      double[] w = new double[A.length];
      double wms=0.0F, ws=0.0F;
      for(int i=0; i<A.length; i++) {
         w[i] = 1.0F / (1.0F + Math.pow((m-A[i])/m,2));
         wms += w[i]*A[i];
         ws  += w[i];
         }
      double M=wms/ws;
      wms = 0.0F;
      for(int i=0; i<A.length; i++)
         wms += w[i]*Math.pow(M-A[i],2);
      double S=Math.sqrt(wms/ws);
      wms = 0.0F;
      ws  = 0.0F;
      for(int i=0; i<A.length; i++) {
         w[i] = w[i]*(((S==0)||(Math.abs(A[i]-M)<(1*S)))?1F:0F);
         wms += w[i]*A[i];
         ws  += w[i];
         }
      return(wms/ws);
      }

   public static double[] theilsen(double[][] arr) {
      return theilsen(arr[0], arr[1]);
      }
   public static double[] theilsen(double[] x, double[] y) {
      if (x.length < 2) return new double[]{Double.NaN,Double.NaN};
      int nf = x.length;

      int np=0;
      for(int i=0; i<(nf-1); i++)
         np+=(nf-i-1);

      double[] yp = new double[np];
      for(int i=0,k=0; i<(nf-1); i++)
         for(int j=i+1; j<nf; j++,k++)
            yp[k] =((y[i]-y[j])/(x[i]-x[j]));
      double M = center(yp);

      double[] xp = new double[np];
      for(int i=0,k=0; i<(nf-1); i++)
         for(int j=i+1; j<nf; j++,k++)
            xp[k] = y[i] - M*x[i];
      double B = center(xp);
 
      double[] mb = {M,B};
      return(mb);
      }
   public static double theilsenB(double[] x, double[] y, double M) {
      if (x.length == 1)
         return y[0] - M*x[0];

      int nf = x.length;
      int np=0;
      for(int i=0; i<(nf-1); i++)
         np+=(nf-i-1);

      double[] xp = new double[np];
      for(int i=0,k=0; i<(nf-1); i++)
         for(int j=i+1; j<nf; j++,k++)
            xp[k] = y[i] - M*x[i];
      return center(xp);
      }

   public static double centerB(double[] x, double[] y, double M) {
      double[] B = new double[x.length];
      for(int i=0; i<x.length; i++)
         B[i] = y[i]-M*x[i];
      return center(B);
      }
   public static double centerM(double[] x, double[] y) {
      double[] M = new double[x.length-1];
      for(int i=1; i<x.length; i++)
         M[i-1] = y[i]/x[i];
      return center(M);
      }

   public static double[] linreg(double[] x, double[] y) {
      double xbar = 0; for(double v : x) xbar += v; xbar /= x.length;
      double ybar = 0; for(double v : y) ybar += v; ybar /= y.length;

      double xxbar = 0, yybar = 0, xybar = 0;
      for(int i=0; i<x.length; i++) {
         double dx = (x[i]-xbar), dy = (y[i]-ybar);
         xxbar += dx*dx;
         yybar += dy*dy;
         xybar += dx*dy;
         }

      double m = xybar/xxbar;
      double b = ybar-m*xbar;
      
      double ssr = 0;
      for(int i=0; i<x.length; i++) {
         double Yi = m*x[i]+b;
         ssr += (y[i] - Yi)*(y[i] - Yi);
         }
      double rs = 1.0 - (ssr/yybar);

      double rxy = xybar/Math.sqrt(xxbar*yybar);

      return new double[]{m,b,rs,rxy};
      }

}
